"""ETL Helper script to demonstrate creating an API to a database using aiohttp."""
import asyncio
import json
import logging
import datetime as dt

import aiohttp
import etlhelper as etl

# import DbParams
from db import ORACLE_DB

logger = logging.getLogger("copy_sensors_async")

# SQL query to get data from Oracle
SELECT_SENSORS = """
    SELECT CODE, DESCRIPTION
    FROM BGS.DIC_SEN_SENSOR
    WHERE date_updated BETWEEN :startdate AND :enddate
    ORDER BY date_updated
    """

# URL of API we want to send data to
BASE_URL = "http://localhost:9200/"
# Headers to tell the API we are sending data in JSON format
HEADERS = {"Content-Type": "application/json"}


def copy_sensors(startdate: dt.datetime, enddate: dt.datetime) -> None:
    """Read sensors from Oracle and post to REST API.
       Requires startdate amd enddate to filter to rows changed in a certain time period.
    """
    logger.info("Copying sensors with timestamps from %s to %s",
                startdate.isoformat(), enddate.isoformat())
    row_count = 0

    # Connect using the imported DbParams
    with ORACLE_DB.connect("ORACLE_PASSWORD") as conn:
        # chunks is a generator that yields lists of dictionaries
        # passing in our select query, connection object, bind variable parameters and custom transform function
        chunks = etl.iter_chunks(
            SELECT_SENSORS,
            conn,
            parameters={"startdate": startdate, "enddate": enddate},
            transform=transform_sensors,
        )

        # for each chunk of rows, synchronously post them to API
        for chunk in chunks:
            result = asyncio.run(post_chunk(chunk))
            row_count += len(result)
            logger.info("%s items transferred", row_count)

    logger.info("Transfer complete")


def transform_sensors(chunk: list[tuple]) -> list[tuple]:
    """Transform rows to dictionaries suitable for converting to JSON."""
    new_chunk = []

    for row in chunk:
        new_row = {
            "sample_code": row.CODE,
            "description": row.DESCRIPTION,
            "metadata": {
                "source": "ORACLE_DB",  # fixed value
                "transferred_at": dt.datetime.now().isoformat(),  # dynamic value
                }
            }
        logger.debug(new_row)
        new_chunk.append(new_row)

    return new_chunk


async def post_chunk(chunk: list[tuple]) -> list:
    """Post multiple items to API asynchronously."""
    # initialize aiohttp session
    async with aiohttp.ClientSession() as session:
        # Build list of tasks
        tasks = []
        # add each row to list of tasks for aiohttp to execute
        for item in chunk:
            # a task is the instance of a function being executed with distinct arguments
            #   in this case, the post_one function with argument of a dictionary representing a row of data
            tasks.append(post_one(item, session))

        # Process tasks in parallel.  An exception in any will be raised.
        result = await asyncio.gather(*tasks)

    return result


async def post_one(item: tuple, session: aiohttp.ClientSession) -> int:
    """Post a single item to API using existing aiohttp Session."""
    # Post the item
    response = await session.post(
        BASE_URL + "sensors/_doc",
        headers=HEADERS,
        # convert python dict to json object
        data=json.dumps(item),
    )

    # Log responses before throwing errors because error info is not included
    # in generated Exceptions and so cannot otherwise be seen for debugging.
    if response.status >= 400:
        response_text = await response.text()
        logger.error(
            "The following item failed: %s\nError message:\n(%s)",
            item,
            response_text,
        )
        await response.raise_for_status()

    return response.status


if __name__ == "__main__":
    # Configure logging
    handler = logging.StreamHandler()
    formatter = logging.Formatter("%(asctime)s %(levelname)-8s %(message)s")
    handler.setFormatter(formatter)
    logger.setLevel(logging.INFO)
    logger.addHandler(handler)

    # Copy data that was updated between 1 January 2000 to 00:00:00 today
    today = dt.datetime.combine(dt.date.today(), dt.time.min)
    copy_sensors(dt.datetime(2000, 1, 1), today)



"""Tests for etl copy functions.  This includes application of transforms.
These are run against PostgreSQL."""
# pylint: disable=unused-argument, missing-docstring
from datetime import (
    datetime,
    date,
)
from typing import (
    Iterable,
    Any,
)

import pytest

from etlhelper import (
    copy_rows,
    copy_table_rows,
    execute,
    fetchall,
    iter_rows,
    load,
)
from etlhelper.row_factories import (
    dict_row_factory,
    namedtuple_row_factory,
)
from etlhelper.exceptions import ETLHelperBadIdentifierError


def test_copy_rows_happy_path(pgtestdb_conn, pgtestdb_test_tables,
                              pgtestdb_insert_sql, test_table_data_namedtuple):
    # Arrange and act
    select_sql = "SELECT * FROM src"
    insert_sql = pgtestdb_insert_sql.replace('src', 'dest')
    processed, failed = copy_rows(select_sql, pgtestdb_conn, insert_sql, pgtestdb_conn,
                                  row_factory=namedtuple_row_factory)

    # Assert
    assert processed == len(test_table_data_namedtuple)
    assert failed == 0

    sql = "SELECT * FROM dest"
    result = iter_rows(sql, pgtestdb_conn, row_factory=namedtuple_row_factory)
    assert list(result) == test_table_data_namedtuple


def test_copy_table_rows_happy_path(pgtestdb_conn, pgtestdb_test_tables, test_table_data_dict):
    # Arrange and act
    processed, failed = copy_table_rows('src', pgtestdb_conn, pgtestdb_conn, target='dest')

    # Assert
    assert processed == len(test_table_data_dict)
    assert failed == 0

    sql = "SELECT * FROM dest"
    result = fetchall(sql, pgtestdb_conn)
    assert result == test_table_data_dict


def test_copy_table_rows_on_error(pgtestdb_test_tables, pgtestdb_conn,
                                  test_table_data_dict):
    # Arrange
    duplicate_id_row_sql = """
       INSERT INTO dest (id, value)
       VALUES (1, 1.234)""".strip()
    execute(duplicate_id_row_sql, pgtestdb_conn)

    # Act
    errors = []
    processed, failed = copy_table_rows('src', pgtestdb_conn, pgtestdb_conn, target='dest',
                                        on_error=errors.extend)

    # Assert
    assert processed == len(test_table_data_dict)
    assert failed == len(errors)

    sql = "SELECT * FROM dest"
    result = fetchall(sql, pgtestdb_conn)

    # Check that first row was caught as error
    row, exception = errors[0]
    assert row["id"] == 1
    assert "unique" in str(exception).lower()

    # Check that other rows were inserted correctly
    assert result[1:] == test_table_data_dict[1:]


def test_copy_table_rows_bad_table(pgtestdb_conn, pgtestdb_test_tables):
    # Arrange and act
    with pytest.raises(ETLHelperBadIdentifierError):
        copy_table_rows('bad; sql', pgtestdb_conn, pgtestdb_conn, target='dest')


def transform_return_list(rows):
    """
    Simple transform function that changes size and number of rows.
    """
    return [(row.id, row.value) for row in rows if row.id > 1]


def transform_yield_new_rows(rows):
    """
    Simple transform function that changes size and number of rows.
    """
    for row in rows:
        new_row = (row.id, row.value)
        if row.id > 1:
            yield new_row


def transform_return_generator(rows):
    """
    Simple transform function that changes size and number of rows.
    """
    return ((row.id, row.value) for row in rows if row.id > 1)


def transform_return_new_list(rows):
    """
    Simple transform function that changes size and number of rows
    using traditional for loop syntax.
    """
    transformed = []
    for row in rows:
        if row.id > 1:
            new_row = (row.id, row.value)
            transformed.append(new_row)

    return transformed


@pytest.mark.parametrize('my_transform',
                         [transform_return_list, transform_return_generator,
                          transform_return_new_list, transform_yield_new_rows])
def test_copy_rows_transform(pgtestdb_conn, pgtestdb_test_tables, my_transform):
    # Arrange
    select_sql = "SELECT * FROM src"
    insert_sql = "INSERT INTO dest (id, value) VALUES (%s, %s)"

    expected = [(2, 2.234, 'default', None, None, None),
                (3, 2.234, 'default', None, None, None)]

    # Act
    copy_rows(select_sql, pgtestdb_conn, insert_sql, pgtestdb_conn,
              transform=my_transform, row_factory=namedtuple_row_factory)

    # Assert
    sql = "SELECT * FROM dest"
    result = iter_rows(sql, pgtestdb_conn, row_factory=namedtuple_row_factory)
    assert list(result) == expected


def transform_yield_modified_dict(chunk):
    """Add 1 thousand to id, replace text with UPPER CASE and replace newlines
    from utf8_text."""
    for row in chunk:
        row['id'] += 1000
        row['simple_text'] = row['simple_text'].upper()
        row['utf8_text'] = row['utf8_text'].replace('\n', ' ')
        yield row


def test_fetchall_with_modify_dict(pgtestdb_conn, pgtestdb_test_tables):
    # Arrange
    select_sql = "SELECT * FROM src LIMIT 1"
    expected = [{
        'date_time': datetime(2018, 12, 7, 13, 1, 59),
        'day': date(2018, 12, 7),
        'id': 1001,
        'simple_text': 'TEXT',
        'utf8_text': 'Öæ° z',
        'value': 1.234
    }]

    # Act
    result = fetchall(select_sql, pgtestdb_conn, row_factory=dict_row_factory,
                      transform=transform_yield_modified_dict)

    # Assert
    assert result == expected


def test_copy_rows_with_dict_row_factory(pgtestdb_conn, pgtestdb_test_tables):
    # Arrange
    select_sql = "SELECT * FROM src"
    insert_sql = """
        INSERT INTO dest (id, value, simple_text, utf8_text, day, date_time)
        VALUES (
            %(id)s, %(value)s, %(simple_text)s,
            %(utf8_text)s, %(day)s, %(date_time)s
        );
    """

    expected = [(1001, 1.234, 'TEXT', 'Öæ° z', date(2018, 12, 7), datetime(2018, 12, 7, 13, 1, 59)),
                (1002, 2.234, 'TEXT', 'Öæ° z', date(2018, 12, 8), datetime(2018, 12, 8, 13, 1, 59)),
                (1003, 2.234, 'TEXT', 'Öæ° z', date(2018, 12, 9), datetime(2018, 12, 9, 13, 1, 59))]

    # Act
    copy_rows(select_sql, pgtestdb_conn, insert_sql, pgtestdb_conn,
              transform=transform_yield_modified_dict)

    # Assert
    sql = "SELECT * FROM dest"
    result = fetchall(sql, pgtestdb_conn, row_factory=namedtuple_row_factory)
    assert result == expected


def convert_namedtuples_dicts(rows):
    """
    Convert a list of namedtuples to dictionaries.
    This function is used to prepare data.
    """
    return [row._asdict() for row in rows]


def transform_namedtuple(rows: list[tuple[Any]]) -> Iterable[tuple[Any]]:
    """
    Simple transform function for a list of namedtuples which:
        - Adds 1000 to each 'id' value
        - Converts the text in the column 'simple_text' to UPPER
    """
    for row in rows:
        row = row._replace(id=row.id + 1000)
        row = row._replace(simple_text=row.simple_text.upper())
        yield row


def transform_dict(rows: list[dict[str, Any]]) -> Iterable[dict[str, Any]]:
    """
    Simple transform function for a list of dictionaries which:
        - Adds 1000 to each 'id' value
        - Converts the text in the column 'upper_text' to UPPER
    """
    for row in rows:
        row["id"] += 1000
        row["simple_text"] = row["simple_text"].upper()

        yield row


@pytest.mark.parametrize(
    ["test_data_conversion", "my_transform", "expected"],
    [
        (
            None,
            transform_namedtuple,
            [
                # pre-loaded data
                (1, 1.234, 'text', 'Öæ°\nz', date(2018, 12, 7), datetime(2018, 12, 7, 13, 1, 59)),
                (2, 2.234, 'text', 'Öæ°\nz', date(2018, 12, 8), datetime(2018, 12, 8, 13, 1, 59)),
                (3, 2.234, 'text', 'Öæ°\nz', date(2018, 12, 9), datetime(2018, 12, 9, 13, 1, 59)),
                # newly loaded transformed data
                (1001, 1.234, 'TEXT', 'Öæ°\nz', date(2018, 12, 7), datetime(2018, 12, 7, 13, 1, 59)),
                (1002, 2.234, 'TEXT', 'Öæ°\nz', date(2018, 12, 8), datetime(2018, 12, 8, 13, 1, 59)),
                (1003, 2.234, 'TEXT', 'Öæ°\nz', date(2018, 12, 9), datetime(2018, 12, 9, 13, 1, 59)),
            ],
        ),
        (
            convert_namedtuples_dicts,
            transform_dict,
            [
                # pre-loaded data
                (1, 1.234, 'text', 'Öæ°\nz', date(2018, 12, 7), datetime(2018, 12, 7, 13, 1, 59)),
                (2, 2.234, 'text', 'Öæ°\nz', date(2018, 12, 8), datetime(2018, 12, 8, 13, 1, 59)),
                (3, 2.234, 'text', 'Öæ°\nz', date(2018, 12, 9), datetime(2018, 12, 9, 13, 1, 59)),
                # newly loaded transformed data
                (1001, 1.234, 'TEXT', 'Öæ°\nz', date(2018, 12, 7), datetime(2018, 12, 7, 13, 1, 59)),
                (1002, 2.234, 'TEXT', 'Öæ°\nz', date(2018, 12, 8), datetime(2018, 12, 8, 13, 1, 59)),
                (1003, 2.234, 'TEXT', 'Öæ°\nz', date(2018, 12, 9), datetime(2018, 12, 9, 13, 1, 59)),
            ],
        ),
    ],
)
def test_load_transform(
    pgtestdb_conn,
    pgtestdb_test_tables,
    test_table_data_namedtuple,
    test_data_conversion,
    my_transform,
    expected,
):
    # Arrange
    if test_data_conversion:
        test_table_data_namedtuple = test_data_conversion(test_table_data_namedtuple)

    # Act
    processed, failed = load(
        table="src",
        conn=pgtestdb_conn,
        rows=test_table_data_namedtuple,
        transform=my_transform,
    )

    sql = "SELECT * FROM src"
    result = fetchall(sql, pgtestdb_conn, row_factory=namedtuple_row_factory)

    # Assert
    for i, row in enumerate(result):
        assert row == expected[i]



"""
Tests for row factory methods.
"""
import sqlite3
from unittest import mock

import pytest

from etlhelper.row_factories import namedtuple_row_factory


FAKE_ROWS = [(1, 2, 3), (4, 5, 6)]


def test_valid_field_names(mock_cursor):
    mock_cursor.description = (('id', None, None, None, None, None, None),
                               ('name', None, None, None, None, None, None),
                               ('desc', None, None, None, None, None, None))

    create_row = namedtuple_row_factory(mock_cursor)
    rows = [create_row(row) for row in mock_cursor.fetchall()]

    assert rows[0]._fields == ("id", "name", "desc")
    assert rows == FAKE_ROWS


def test_invalid_field_names(mock_cursor):
    mock_cursor.description = (('id', None, None, None, None, None, None),
                               ('count(*)', None, None, None, None, None, None),
                               ('spaced column', None, None, None, None, None, None))

    with pytest.warns(UserWarning) as warn:
        create_row = namedtuple_row_factory(mock_cursor)
        assert len(warn) == 2
        assert (warn[1].message.args[0]
                == 'count(*) was renamed to _1\nspaced column was renamed to _2')

    rows = [create_row(row) for row in mock_cursor.fetchall()]

    assert rows[0]._fields == ('id', '_1', '_2')
    assert rows == FAKE_ROWS


@pytest.fixture(scope='function')
def mock_cursor():
    # Test uses sqlite3 cursor, but all dbapi compliant cursors will be the same.
    mock_cursor = mock.MagicMock(sqlite3.Cursor)
    mock_cursor.fetchall.return_value = FAKE_ROWS
    return mock_cursor



"""ETL Helper script to demonstrate using executemany with a positional placeholder query."""
import sqlite3
import etlhelper as etl

db_file = "igneous_rocks.db"

# Positional placeholders for data in tuple format
insert_sql = """
    INSERT INTO igneous_rocks (name, grain_size, updated_at)
    VALUES (?, UPPER(?), datetime('now'))
"""

rows = [("basalt", "fine"), ("granite", "coarse")]

with sqlite3.connect(db_file) as conn:
    # Note that table must already exist
    processed, failed = etl.executemany(insert_sql, conn, rows)



"""
Functions for transferring data in and out of databases.
"""
import logging
import re
from copy import deepcopy
from itertools import (
    zip_longest,
    chain,
)
from typing import (
    Any,
    Callable,
    Collection,
    Iterable,
    Iterator,
    NamedTuple,
    Optional,
    Union,
)

from etlhelper.abort import (
    raise_for_abort,
    clear_abort_event,
)
from etlhelper.db_helper_factory import DB_HELPER_FACTORY
from etlhelper.exceptions import (
    ETLHelperBadIdentifierError,
    ETLHelperExtractError,
    ETLHelperInsertError,
    ETLHelperQueryError,
)
from etlhelper.row_factories import dict_row_factory
from etlhelper.types import (
    Connection,
    Row,
    Chunk,
)

logger = logging.getLogger('etlhelper')
CHUNKSIZE = 5000


class FailedRow(NamedTuple):
    row: Row
    exception: Exception


def iter_chunks(
        select_query: str,
        conn: Connection,
        parameters: tuple = (),
        row_factory: Callable = dict_row_factory,
        transform: Optional[Callable[[Chunk], Chunk]] = None,
        chunk_size: int = CHUNKSIZE
        ) -> Iterator[Chunk]:
    """
    Run SQL query against connection and return iterator object to loop over
    results in batches of chunksize (default 5000).

    The row_factory changes the output format of the results.  Other row
    factories e.g. dict_row_factory are available.

    The transform function is applied to chunks of data as they are extracted
    from the database.

    All data extraction functions call this function, directly or indirectly.

    :param select_query: SQL query to execute
    :param conn: dbapi connection
    :param parameters: bind variables to insert in the query
    :param row_factory: function that accepts a cursor and returns a function
                        for parsing each row
    :param transform: function that accepts a list of rows and
                      returns an list of rows (possibly of different shape)
    :param chunk_size: size of chunks to group data by
    :return: generator returning a list of objects which each
             represent a row of data using the given row_factory
    :raises ETLHelperExtractError: if SQL raises an error
    """
    logger.info("Fetching rows (chunk_size=%s)", chunk_size)
    logger.debug(f"Fetching:\n\n{select_query}\n\nwith parameters:\n\n"
                 f"{parameters}\n\nagainst:\n\n{conn}")
    clear_abort_event()

    helper = DB_HELPER_FACTORY.from_conn(conn)
    with helper.cursor(conn) as cursor:

        # Run query
        try:
            cursor.execute(select_query, parameters)
        except helper.sql_exceptions as exc:
            # Even though we haven't modified data, we have to rollback to
            # clear the failed transaction before any others can be started.
            conn.rollback()
            msg = (f"SQL query raised an error.\n\n{select_query}\n\n"
                   f"Required paramstyle: {helper.paramstyle}\n\n{exc}\n")
            raise ETLHelperExtractError(msg)

        # Set row factory
        create_row = row_factory(cursor)

        # Parse results
        first_pass = True
        while True:
            raise_for_abort("abort_etlhelper_threads() called during iter_chunks")

            rows = cursor.fetchmany(chunk_size)

            # No more rows to process
            if not rows:
                if first_pass:
                    msg = "No rows returned"
                else:
                    if cursor.rowcount == -1:
                        # SQLite3 drive doesn't support row count (always -1)
                        msg = "All rows returned"
                    else:
                        msg = f"{cursor.rowcount} rows returned"
                logger.info(msg)

                # Close the active transaction
                conn.commit()
                return

            # Apply row_factory
            rows = (create_row(row) for row in rows)

            # Apply transform
            if transform:
                rows = transform(rows)

            # Return data
            yield rows
            first_pass = False


def iter_rows(
        select_query: str,
        conn: Connection,
        parameters: tuple = (),
        row_factory: Callable = dict_row_factory,
        transform: Optional[Callable[[Chunk], Chunk]] = None,
        chunk_size: int = CHUNKSIZE
        ) -> Iterator[Row]:
    """
    Run SQL query against connection and return iterator object to loop over
    results, row-by-row.

    :param select_query: SQL query to execute
    :param conn: dbapi connection
    :param parameters: bind variables to insert in the query
    :param row_factory: function that accepts a cursor and returns a function
                        for parsing each row
    :param transform: function that accepts a list of rows and
                      returns an list of rows (possibly of different shape)
    :param chunk_size: size of chunks to group data by
    :return: generator returning a list of objects which each
             represent a row of data using the given row_factory
    """
    for chunk in iter_chunks(select_query, conn, row_factory=row_factory,
                             parameters=parameters, transform=transform,
                             chunk_size=chunk_size):
        for row in chunk:
            yield row


def fetchone(
        select_query: str,
        conn: Connection,
        parameters: tuple = (),
        row_factory: Callable = dict_row_factory,
        transform: Optional[Callable[[Chunk], Chunk]] = None,
        chunk_size: int = 1
        ) -> Optional[Row]:
    """
    Get first result of query.  See iter_rows for details.  Note: iter_rows is
    recommended for looping over rows individually.

    :param select_query: SQL query to execute
    :param conn: dbapi connection
    :param parameters: bind variables to insert in the query
    :param row_factory: function that accepts a cursor and returns a function
                        for parsing each row
    :param transform: function that accepts a list of rows and
                      returns an list of rows (possibly of different shape)
    :param chunk_size: size of chunks to group data by
    :return: None or a row of data using the given row_factory
    """
    try:
        result = next(iter_rows(select_query, conn, row_factory=row_factory,
                                parameters=parameters, transform=transform,
                                chunk_size=chunk_size))
    except StopIteration:
        result = None
    finally:
        # Commit to close the transaction before the iterator has been exhausted
        conn.commit()

    return result


def fetchall(
        select_query: str,
        conn: Connection,
        parameters: tuple = (),
        row_factory: Callable = dict_row_factory,
        transform: Optional[Callable[[Chunk], Chunk]] = None,
        chunk_size: int = CHUNKSIZE
        ) -> Chunk:
    """
    Get all results of query as a list.  See iter_rows for details.

    :param select_query: SQL query to execute
    :param conn: dbapi connection
    :param parameters: bind variables to insert in the query
    :param row_factory: function that accepts a cursor and returns a function
                        for parsing each row
    :param transform: function that accepts a list of rows and
                      returns an list of rows (possibly of different shape)
    :param chunk_size: size of chunks to group data by
    :return: a row of data using the given row_factory
    """
    return list(iter_rows(select_query, conn, row_factory=row_factory,
                          parameters=parameters, transform=transform,
                          chunk_size=chunk_size))


def executemany(
        query: str,
        conn: Connection,
        rows: Iterable[Row],
        transform: Optional[Callable[[Chunk], Chunk]] = None,
        on_error: Optional[Callable[[list[FailedRow]], Any]] = None,
        commit_chunks: bool = True,
        chunk_size: int = CHUNKSIZE,
        ) -> tuple[int, int]:
    """
    Use query to insert/update data from rows to database at conn.  This
    method uses the executemany or execute_batch (PostgreSQL) commands to
    process the data in chunks and avoid creating a new database connection for
    each row.  Row data are passed as parameters into query.

    Default behaviour is to raise an exception in the case of SQL errors such
    as primary key violations.  If the on_error parameter is specified, the
    exception will be caught then then rows of each chunk re-tried individually.
    Further errors will be caught and appended to a list of (row, exception)
    tuples.  on_error is a function that is called at the end of each chunk,
    with the list as the only argument.

    commit_chunks controls if chunks the transaction should be committed after
    each chunk has been inserted.  Committing chunks means that errors during
    a long-running insert do not require all data to be loaded again.  The
    disadvantage is that investigation may be required to determine exactly
    which records have been successfully transferred.

    :param query: SQL insert command with placeholders for data
    :param conn: dbapi connection
    :param rows: an iterable of rows containing data to be inserted/updated
    :param transform: function that accepts a list of rows and
                      returns an list of rows (possibly of different shape)
    :param on_error: function to be applied to failed rows in each chunk
    :param commit_chunks: commit after each chunk has been inserted/updated
    :param chunk_size: size of chunks to group data by
    :return: the number of rows processed and the number of rows failed
    :raises ETLHelperInsertError: if SQL raises an error
    """
    logger.info("Executing many (chunk_size=%s)", chunk_size)
    logger.debug("Executing:\n\n%s\n\nagainst:\n\n%s", query, conn)
    clear_abort_event()

    helper = DB_HELPER_FACTORY.from_conn(conn)
    processed = 0
    failed = 0

    with helper.cursor(conn) as cursor:
        for chunk_with_nones in _chunker(rows, chunk_size):
            raise_for_abort("abort_etlhelper_threads() called during executemany")

            # Chunker pads to whole chunk with None; remove these
            chunk = [row for row in chunk_with_nones if row is not None]

            # Apply transform
            if transform:
                # Materialise chunk as list as `transform` can return generator
                chunk = list(transform(chunk))

            # Show first row as example of data
            if processed == 0:
                logger.debug(f"First row: {chunk[0]}")

            # Run query
            try:
                helper.executemany(cursor, query, chunk)

            except helper.sql_exceptions as exc:
                # Rollback to clear the failed transaction before any others can
                # be started.
                conn.rollback()

                # Collect and process failed rows if on_error function provided
                if on_error:
                    # Temporarily disable logging
                    old_level = logger.level
                    logger.setLevel(logging.ERROR)

                    try:
                        failed_rows = _execute_by_row(query, conn, chunk)
                    finally:
                        # Restore logging
                        logger.setLevel(old_level)

                    failed += len(failed_rows)
                    logger.debug("Calling on_error function on %s failed rows",
                                 failed)
                    on_error(failed_rows)
                else:
                    msg = (f"SQL query raised an error.\n\n{query}\n\n"
                           f"Required paramstyle: {helper.paramstyle}\n\n{exc}\n")
                    raise ETLHelperInsertError(msg)

            processed += len(chunk)
            logger.info(
                '%s rows processed (%s failed)', processed, failed)

            # Commit changes so far
            if commit_chunks:
                conn.commit()

    # Commit changes where not already committed
    if not commit_chunks:
        conn.commit()

    logger.info(f'{processed} rows processed in total')
    return processed, failed


def _execute_by_row(
        query: str,
        conn: Connection,
        chunk: Chunk
        ) -> list[FailedRow]:
    """
    Retry execution of rows individually and return failed rows along with
    their errors. Successful inserts are committed.

    :param query: SQL query with placeholders for data
    :param conn: dbapi connection
    :param chunk: list of rows
    :return: a list failed rows
    """
    failed_rows: list[FailedRow] = []

    for row in chunk:
        try:
            # Use etlhelper execute to isolate transactions
            execute(query, conn, parameters=row)
        except ETLHelperQueryError as exc:
            failed_rows.append(FailedRow(row, exc))

    return failed_rows


def copy_rows(
        select_query: str,
        source_conn: Connection,
        insert_query: str,
        dest_conn: Connection,
        parameters: tuple = (),
        row_factory: Callable = dict_row_factory,
        transform: Optional[Callable[[Chunk], Chunk]] = None,
        on_error: Optional[Callable] = None,
        commit_chunks: bool = True,
        chunk_size: int = CHUNKSIZE,
        ) -> tuple[int, int]:
    """
    Copy rows from source_conn to dest_conn.  select_query and insert_query
    specify the data to be transferred.

    Note: ODBC driver requires separate connections for source_conn and
    dest_conn, even if they represent the same database.

    Geometry columns from Oracle spatial should be selected with:
      SDO_UTIL.TO_WKTGEOMETRY(shape_bng) AS geom_wkt
    and inserted into PostGIS with:
      ST_GeomFromText(%s, 27700)

    Default behaviour is to raise an exception in the case of SQL errors such
    as primary key violations.  If the on_error parameter is specified, the
    exception will be caught then then rows of each chunk re-tried individually.
    Further errors will be caught and appended to a list of (row, exception)
    tuples.  on_error is a function that is called at the end of each chunk,
    with the list as the only argument.

    :param select_query: SQL query to select data
    :param source_conn: dbapi connection
    :param insert_query: SQL query to insert data
    :param dest_conn: dbapi connection
    :param parameters: bind variables to insert in the select query
    :param row_factory: function that accepts a cursor and returns a function
                        for parsing each row
    :param transform: function that accepts a list of rows and
                      returns an list of rows (possibly of different shape)
    :param on_error: function to be applied to failed rows in each chunk
    :param commit_chunks: commit after each chunk (see executemany)
    :param chunk_size: size of chunks to group data by
    :return: the number of rows processed and the number of rows failed
    """
    rows_generator = iter_rows(select_query, source_conn,
                               parameters=parameters, row_factory=row_factory,
                               transform=transform, chunk_size=chunk_size)
    processed, failed = executemany(insert_query, dest_conn,
                                    rows_generator,
                                    on_error=on_error,
                                    commit_chunks=commit_chunks,
                                    chunk_size=chunk_size)
    return processed, failed


def execute(
        query: str,
        conn: Connection,
        parameters: Collection[Any] = ()
        ) -> None:
    """
    Run SQL query against connection.

    :param query: SQL query to execute
    :param conn: dbapi connection
    :param parameters: bind variables to insert in the query
    :raises ETLHelperQueryError: if SQL raises an error
    """
    logger.info("Executing query")
    logger.debug(f"Executing:\n\n{query}\n\nwith parameters:\n\n"
                 f"{parameters}\n\nagainst:\n\n{conn}")

    helper = DB_HELPER_FACTORY.from_conn(conn)
    with helper.cursor(conn) as cursor:
        # Run query
        try:
            cursor.execute(query, parameters)
            conn.commit()
        except helper.sql_exceptions as exc:
            # Even though we haven't modified data, we have to rollback to
            # clear the failed transaction before any others can be started.
            conn.rollback()
            msg = (f"SQL query raised an error.\n\n{query}\n\n"
                   f"Required paramstyle: {helper.paramstyle}\n\n{exc}\n")
            raise ETLHelperQueryError(msg)


def copy_table_rows(
        table: str,
        source_conn: Connection,
        dest_conn: Connection,
        target: Optional[str] = None,
        row_factory: Callable = dict_row_factory,
        transform: Optional[Callable[[Chunk], Chunk]] = None,
        on_error: Optional[Callable] = None,
        commit_chunks: bool = True,
        chunk_size: int = CHUNKSIZE
        ) -> tuple[int, int]:
    """
    Copy rows from 'table' in source_conn to same or target table in dest_conn.
    This is a simple copy of all columns and rows using `load` to insert data.
    It is possible to apply a transform e.g. to change the case of table names.
    For more control, use `copy_rows`.

    Note: ODBC driver requires separate connections for source_conn and
    dest_conn, even if they represent the same database.

    Default behaviour is to raise an exception in the case of SQL errors such
    as primary key violations.  If the on_error parameter is specified, the
    exception will be caught then then rows of each chunk re-tried individually.
    Further errors will be caught and appended to a list of (row, exception)
    tuples.  on_error is a function that is called at the end of each chunk,
    with the list as the only argument.

    :param table: name of table
    :param source_conn: dbapi connection
    :param dest_conn: dbapi connection
    :param target: name of target table, if different from source
    :param row_factory: function that accepts a cursor and returns a function
                        for parsing each row
    :param transform: function that accepts a list of rows and
                      returns an list of rows (possibly of different shape)
    :param on_error: function to be applied to failed rows in each chunk
    :param commit_chunks: commit after each chunk (see executemany)
    :param chunk_size: size of chunks to group data by
    :return: the number of rows processed and the number of rows failed
    """
    validate_identifier(table)

    select_query = f"SELECT * FROM {table}"
    if not target:
        target = table

    rows_generator = iter_rows(select_query, source_conn, row_factory=row_factory,
                               transform=transform, chunk_size=chunk_size)
    processed, failed = load(target, dest_conn, rows_generator, on_error=on_error,
                             commit_chunks=commit_chunks, chunk_size=chunk_size)
    return processed, failed


def load(
        table: str,
        conn: Connection,
        rows: Iterable[Row],
        transform: Optional[Callable[[Chunk], Chunk]] = None,
        on_error: Optional[Callable] = None,
        commit_chunks: bool = True,
        chunk_size: int = CHUNKSIZE,
        ) -> tuple[int, int]:
    """
    Load data from iterable of named tuples or dictionaries into pre-existing
    table in database on conn.

    Default behaviour is to raise an exception in the case of SQL errors such
    as primary key violations.  If the on_error parameter is specified, the
    exception will be caught then then rows of each chunk re-tried individually.
    Further errors will be caught and appended to a list of (row, exception)
    tuples.  on_error is a function that is called at the end of each chunk,
    with the list as the only argument.

    :param table: name of table
    :param conn: dbapi connection
    :param rows: iterable of named tuples or dictionaries of data
    :param transform: function that accepts a list of rows and
                      returns an list of rows (possibly of different shape)
    :param on_error: function to be applied to failed rows in each chunk
    :param commit_chunks: commit after each chunk (see executemany)
    :param chunk_size: size of chunks to group data by
    :return: the number of rows processed and the number of rows failed
    """
    # Return early if rows is empty
    if not rows:
        return 0, 0

    # Get first row without losing it from row iteration, returning early if
    # the generator was empty.
    try:
        rows = iter(rows)
        first_row = next(rows)  # This line throws an exception if empty
        rows = chain([first_row], rows)

        if transform:
            # next(iter(var)) is equivalent to var[0] but also works for generators,
            # which may be returned by the transform.
            # We need a deepcopy to stop the first row being transformed twice.
            first_row_transformed = next(iter(transform([deepcopy(first_row)])))
        else:
            first_row_transformed = first_row

    except StopIteration:
        return 0, 0

    # Generate insert query
    query = generate_insert_sql(table, first_row_transformed, conn)

    # Insert data
    processed, failed = executemany(
        query,
        conn,
        rows,
        transform=transform,
        on_error=on_error,
        commit_chunks=commit_chunks,
        chunk_size=chunk_size,
    )
    return processed, failed


def generate_insert_sql(
        table: str,
        row: Row,
        conn: Connection
        ) -> str:
    """Generate insert SQL for table, getting column names from row and the
    Generate insert SQL for table, getting column names from row and the
    placeholder style from the connection.  `row` is either a namedtuple or
    a dictionary.

    :param table: name of table
    :param row: a single row as a namedtuple or dict
    :param conn: dbapi connection
    :return: SQL statement to insert data into the given table
    :raises ETLHelperInsertError: if 'row' is not a namedtuple or a dict,
                                  or if the database connection encounters a
                                  parameter error
    """
    helper = DB_HELPER_FACTORY.from_conn(conn)
    paramstyles = {
        "qmark": "?",
        "numeric": ":{number}",
        "named": ":{name}",
        "format": "%s",
        "pyformat": "%({name})s"
    }

    # Namedtuples use a query with positional placeholders
    if hasattr(row, '_asdict'):
        paramstyle = helper.positional_paramstyle

        # Convert namedtuple to dictionary to easily access keys
        row_dict = row._asdict()
        columns = row_dict.keys()
        if paramstyle == "numeric":
            placeholders = [paramstyles[paramstyle].format(number=i + 1)
                            for i in range(len(columns))]
        else:
            placeholders = [paramstyles[paramstyle]] * len(columns)

    # Dictionaries use a query with named placeholders
    elif hasattr(row, 'keys'):
        paramstyle = helper.named_paramstyle
        if not paramstyle:
            msg = (f"Database connection ({str(conn.__class__)}) doesn't support named parameters.  "
                   "Pass data as namedtuples instead.")
            raise ETLHelperInsertError(msg)

        columns = row.keys()
        placeholders = [paramstyles[paramstyle].format(name=c) for c in columns]

    else:
        msg = f"Row is not a dictionary or namedtuple ({type(row)})"
        raise ETLHelperInsertError(msg)

    # Validate identifiers to prevent malicious code injection
    for identifier in (table, *columns):
        validate_identifier(identifier)

    # Generate insert statement
    sql = f"INSERT INTO {table} ({', '.join(columns)}) VALUES ({', '.join(placeholders)})"

    return sql


def validate_identifier(identifier: str) -> None:
    """
    Validate characters used in identifier e.g. table or column name.
    Identifiers must comprise alpha-numeric characters, plus `_` or `$` and
    cannot start with `$`, or numbers.

    :param identifier: a database identifier
    :raises ETLHelperBadIdentifierError: if the 'identifier' contains invalid
                                         characters
    """
    # Identifier rules are based on PostgreSQL specifications, defined here:
    # https://www.postgresql.org/docs/current/sql-syntax-lexical.html#SQL-SYNTAX-IDENTIFIERS

    # `\w` represents all alphanumeric characters (including unicode) plus `_`
    # `(?![0-9])` is a "negative-lookahead assertion" to remove numbers from
    # the match for the first character.
    # The regex comprises two very similar groups.  The first is optional and
    # ends with a dot.  This represents the schema name.
    regex = re.compile(r"((?![0-9])[\w][\w$]*\.?)?((?![0-9])[\w][\w$]*)$")

    if not regex.match(identifier):
        msg = f"'{identifier}' contains invalid characters."
        raise ETLHelperBadIdentifierError(msg)


def _chunker(
        iterable: Iterable[Row],
        n_chunks: int,
        ) -> Iterator[tuple[Union[Row, None], ...]]:
    """Collect data into fixed-length chunks or blocks.
    Code from recipe at https://docs.python.org/3.6/library/itertools.html

    :param iterable: an iterable object
    :param n_chunks: the number of values in each chunk
    :return: generator returning tuples of rows, of length n_chunks,
             where empty values are filled using None
    """
    # _chunker((A,B,C,D,E,F,G), 3) --> (A,B,C) (D,E,F) (G,None,None)
    args = [iter(iterable)] * n_chunks
    return zip_longest(*args, fillvalue=None)



"""Test for etl load functions.  Data loading is carried out using
the executemany call.  These are run against PostgreSQL."""
# pylint: disable=unused-argument, missing-docstring
import re
from collections import namedtuple
from unittest.mock import (
    Mock,
    ANY,
    sentinel,
)

import pytest

import etlhelper.etl as etlhelper_etl
from etlhelper import (
    executemany,
    generate_insert_sql,
    fetchall,
    iter_rows,
    load,
)
from etlhelper.row_factories import namedtuple_row_factory
from etlhelper.etl import ETLHelperInsertError
from etlhelper.exceptions import ETLHelperBadIdentifierError


@pytest.mark.parametrize('commit_chunks', [True, False])
def test_insert_rows_happy_path(pgtestdb_conn, pgtestdb_test_tables,
                                pgtestdb_insert_sql, test_table_data_namedtuple,
                                commit_chunks):
    # Parameterized to ensure success with and without commit_chunks
    # Arrange
    insert_sql = pgtestdb_insert_sql.replace('src', 'dest')

    # Act
    processed, failed = executemany(insert_sql, pgtestdb_conn, test_table_data_namedtuple,
                                    commit_chunks=commit_chunks)

    # Assert
    assert processed == len(test_table_data_namedtuple)
    assert failed == 0

    sql = "SELECT * FROM dest"
    result = fetchall(sql, pgtestdb_conn, row_factory=namedtuple_row_factory)
    assert result == test_table_data_namedtuple


@pytest.mark.parametrize('commit_chunks', [True, False])
def test_insert_rows_on_error(pgtestdb_conn, pgtestdb_test_tables,
                              pgtestdb_insert_sql, test_table_data_namedtuple,
                              commit_chunks):
    # Parameterized to ensure success with and without commit_chunks
    # Arrange
    insert_sql = pgtestdb_insert_sql.replace('src', 'dest')
    # Create duplicated rows to data that will fail primary key check
    duplicated_rows = test_table_data_namedtuple * 2

    # Act
    errors = []
    processed, failed = executemany(insert_sql, pgtestdb_conn, duplicated_rows,
                                    on_error=errors.extend, commit_chunks=commit_chunks)

    # Assert
    assert processed == len(duplicated_rows)
    assert failed == len(errors)

    sql = "SELECT * FROM dest"
    result = fetchall(sql, pgtestdb_conn, row_factory=namedtuple_row_factory)
    assert result == test_table_data_namedtuple

    # Assert full set of failed rows failing unique constraint
    failed_rows, exceptions = zip(*errors)
    assert set(failed_rows) == set(test_table_data_namedtuple)
    assert all(['unique' in str(e).lower() for e in exceptions])


@pytest.mark.parametrize('chunk_size', [1, 2, 3, 4])
def test_insert_rows_chunked(pgtestdb_conn, pgtestdb_test_tables,
                             pgtestdb_insert_sql, test_table_data_namedtuple, monkeypatch,
                             chunk_size):
    # Arrange
    monkeypatch.setattr('etlhelper.etl.CHUNKSIZE', chunk_size)
    insert_sql = pgtestdb_insert_sql.replace('src', 'dest')

    # Act
    executemany(insert_sql, pgtestdb_conn, test_table_data_namedtuple)

    # Assert
    sql = "SELECT * FROM dest"
    result = fetchall(sql, pgtestdb_conn, row_factory=namedtuple_row_factory)
    assert result == test_table_data_namedtuple


def test_insert_rows_no_rows(pgtestdb_conn, pgtestdb_test_tables,
                             pgtestdb_insert_sql):
    # Arrange
    insert_sql = pgtestdb_insert_sql.replace('src', 'dest')

    # Act
    processed, failed = executemany(insert_sql, pgtestdb_conn, [])

    # Assert
    assert processed == 0
    assert failed == 0

    sql = "SELECT * FROM dest"
    result = iter_rows(sql, pgtestdb_conn)
    assert list(result) == []


def test_insert_rows_bad_query(pgtestdb_conn, test_table_data_namedtuple):
    # Arrange
    insert_sql = "INSERT INTO bad_table VALUES (%s, %s, %s, %s, %s, %s)"

    # Act and assert
    with pytest.raises(ETLHelperInsertError):
        executemany(insert_sql, pgtestdb_conn, test_table_data_namedtuple)


def test_load_namedtuples(pgtestdb_conn, pgtestdb_test_tables, test_table_data_namedtuple):
    # Act
    processed, failed = load('dest', pgtestdb_conn, test_table_data_namedtuple)

    # Assert
    assert processed == len(test_table_data_namedtuple)
    assert failed == 0

    sql = "SELECT * FROM dest"
    result = fetchall(sql, pgtestdb_conn, row_factory=namedtuple_row_factory)
    assert result == test_table_data_namedtuple


def test_load_dicts(pgtestdb_conn, pgtestdb_test_tables, test_table_data_dict):
    # Act
    processed, failed = load('dest', pgtestdb_conn, test_table_data_dict)

    # Assert
    assert processed == len(test_table_data_dict)
    assert failed == 0

    sql = "SELECT * FROM dest"
    result = fetchall(sql, pgtestdb_conn)
    assert result == test_table_data_dict


def test_load_namedtuples_with_transform_generator(pgtestdb_conn, pgtestdb_test_tables, test_table_data_namedtuple):
    # Act
    def transform(chunk):
        for row in chunk:
            yield row

    processed, failed = load('dest', pgtestdb_conn, test_table_data_namedtuple, transform=transform)

    # Assert
    assert processed == len(test_table_data_namedtuple)
    assert failed == 0

    sql = "SELECT * FROM dest"
    result = fetchall(sql, pgtestdb_conn, row_factory=namedtuple_row_factory)
    assert result == test_table_data_namedtuple


@pytest.mark.parametrize('null_data', [
    None,
    [],
    (),
    (item for item in ())  # empty generator
    ])
def test_load_no_data(pgtestdb_conn, pgtestdb_test_tables, null_data):
    # Act
    # Function should not crash when data are missing
    processed, failed = load('dest', pgtestdb_conn, null_data)

    assert processed == 0
    assert failed == 0


@pytest.mark.parametrize('chunk_size', [1, 2, 3, 4])
def test_load_namedtuples_chunk_size(pgtestdb_conn, pgtestdb_test_tables,
                                     test_table_data_namedtuple, chunk_size):
    # Act
    load('dest', pgtestdb_conn, test_table_data_namedtuple, chunk_size=chunk_size)

    # Assert
    sql = "SELECT * FROM dest"
    result = fetchall(sql, pgtestdb_conn, row_factory=namedtuple_row_factory)
    assert result == test_table_data_namedtuple


def test_load_parameters_pass_to_executemany(monkeypatch, pgtestdb_conn,
                                             test_table_data_namedtuple):
    # Arrange
    # Patch 'iter_rows' function within etlhelper.etl module
    mock_executemany = Mock()
    mock_executemany.return_value = (0, 0)
    monkeypatch.setattr(etlhelper_etl, 'executemany', mock_executemany)
    # Sentinel items are unique so confirm object that was passed through
    table = 'my_table'
    commit_chunks = sentinel.commit_chunks
    chunk_size = sentinel.chunk_size

    # Act
    load(table, pgtestdb_conn, test_table_data_namedtuple, commit_chunks=commit_chunks,
         chunk_size=chunk_size)

    # Assert
    # load() function writes SQL query
    sql = """
      INSERT INTO my_table (id, value, simple_text, utf8_text, day,
          date_time)
      VALUES (%s, %s, %s, %s, %s, %s)""".strip()
    sql = re.sub(r"\s\s+", " ", sql)  # replace newlines and whitespace

    mock_executemany.assert_called_once_with(
        sql, pgtestdb_conn, ANY,
        transform=None, on_error=None,
        commit_chunks=sentinel.commit_chunks,
        chunk_size=sentinel.chunk_size)


def test_generate_insert_sql_tuple(pgtestdb_conn):
    # Arrange
    data = (1, 'one')

    # Act
    with pytest.raises(ETLHelperInsertError,
                       match="Row is not a dictionary or namedtuple"):
        generate_insert_sql('my_table', data, pgtestdb_conn)


def test_generate_insert_sql_namedtuple(pgtestdb_conn):
    # Arrange
    TwoColumnRow = namedtuple('TwoColumnRow', ('id', 'data'))
    data = TwoColumnRow(1, 'one')
    expected = 'INSERT INTO my_table (id, data) VALUES (%s, %s)'

    # Act
    sql = generate_insert_sql('my_table', data, pgtestdb_conn)

    # Assert
    assert sql == expected


def test_generate_insert_sql_dictionary(pgtestdb_conn):
    # Act
    data = {'id': 1, 'data': 'one'}
    expected = 'INSERT INTO my_table (id, data) VALUES (%(id)s, %(data)s)'

    # Act
    sql = generate_insert_sql('my_table', data, pgtestdb_conn)

    # Assert
    assert sql == expected


def test_generate_insert_sql_None(pgtestdb_conn):
    # Arrange
    data = None

    # Act
    with pytest.raises(ETLHelperInsertError,
                       match="Row is not a dictionary or namedtuple"):
        generate_insert_sql('my_table', data, pgtestdb_conn)


@pytest.mark.parametrize('table, data', [
    ('Bobby; DROP TABLE students', {'id': 1}),  # bad table
    ('my_table', {'Bobby; DROP TABLE students': 1})  # bad column
])
def test_generate_insert_sql_bad_identifers(table, data, pgtestdb_conn):
    # Act
    with pytest.raises(ETLHelperBadIdentifierError):
        generate_insert_sql(table, data, pgtestdb_conn)



"""ETL Helper script to demonstrate DbParams."""
import etlhelper as etl

ORACLEDB = etl.DbParams(
    dbtype="ORACLE",
    host="localhost",
    port=1521,
    dbname="mydata",
    user="oracle_user",
)

POSTGRESDB = etl.DbParams(
    dbtype="PG",
    host="localhost",
    port=5432,
    dbname="mydata",
    user="postgres_user",
)

SQLITEDB = etl.DbParams(
    dbtype="SQLITE",
    filename="/path/to/file.db",
)

MSSQLDB = etl.DbParams(
    dbtype="MSSQL",
    host="localhost",
    port=1433,
    dbname="mydata",
    user="mssql_user",
    odbc_driver="ODBC Driver 17 for SQL Server",
)



"""ETL Helper script to demonstrate iter_rows."""
import sqlite3
import etlhelper as etl

db_file = "igneous_rocks.db"

with sqlite3.connect(db_file) as conn:
    for row in etl.iter_rows("SELECT * FROM igneous_rock", conn):
        print(row)



"""
Database helper for mssql
"""
import warnings
from textwrap import dedent

from etlhelper.db_helpers.db_helper import DbHelper
from etlhelper.exceptions import ETLHelperInsertError


class MSSQLDbHelper(DbHelper):
    """
    MS Sql server helper class
    """
    table_info_query = dedent("""
        SELECT
            column_name as name,
            data_type as type,
            (case when is_nullable = 'NO' then 1 else 0 end) as not_null,
            (case when column_default is not null then 1 else 0 end) as has_default
        FROM INFORMATION_SCHEMA.COLUMNS
        WHERE LOWER(table_name) = LOWER(?)
        AND LOWER(table_schema) LIKE COALESCE(LOWER(?), '%%')
        """).strip()

    def __init__(self):
        super().__init__()
        self.required_params = {'host', 'port', 'dbname', 'user', 'odbc_driver'}
        self.missing_driver_msg = (
            "Could not import pyodbc module required for MS SQL connections.  "
            "See https://github.com/BritishGeologicalSurvey/etlhelper for installation instructions")
        self.named_paramstyle = None  # pyodbc doesn't support named parameters
        self.positional_paramstyle = 'qmark'

        try:
            import pyodbc
            self.sql_exceptions = (pyodbc.DatabaseError,
                                   pyodbc.InterfaceError)
            self.connect_exceptions = (pyodbc.DatabaseError,
                                       pyodbc.InterfaceError)
            self.paramstyle = pyodbc.paramstyle
            self._connect_func = pyodbc.connect
            self.use_fast_executemany = True
            self.trust_server_certificate = False
        except ImportError:
            warnings.warn(self.missing_driver_msg)

    def connect(self, db_params, password_variable=None,
                fast_executemany=True, trust_server_certificate=False, **kwargs):
        self.use_fast_executemany = fast_executemany
        self.trust_server_certificate = trust_server_certificate
        return super().connect(db_params, password_variable, **kwargs)

    def get_connection_string(self, db_params, password_variable):
        """
        Return a connection string
        :param db_params: DbParams
        :param password_variable: str, password
        :return: str
        """
        # Prepare connection string
        password = self.get_password(password_variable)
        conn_str = (f'DRIVER={db_params.odbc_driver};SERVER=tcp:{db_params.host};PORT={db_params.port};'
                    f'DATABASE={db_params.dbname};UID={db_params.user};PWD={password}')
        if self.trust_server_certificate:
            conn_str += ';TrustServerCertificate=yes'
        return conn_str

    def get_sqlalchemy_connection_string(self, db_params, password_variable):
        """
        Returns connection string for SQLAlchemy engine.
        """
        password = self.get_password(password_variable)
        driver = db_params.odbc_driver.replace(" ", "+")
        return (f'mssql+pyodbc://{db_params.user}:{password}@'
                f'{db_params.host}:{db_params.port}/{db_params.dbname}?'
                f'driver={driver}')

    def executemany(self, cursor, query, chunk):
        """
        Try to use fast_executemany for SQL Server if flag in DbParams is set.

        :param cursor: Open database cursor.
        :param query: str, SQL query
        :param chunk: list, Rows of parameters.
        """
        try:
            cursor.fast_executemany = self.use_fast_executemany
            cursor.executemany(query, chunk)
        except MemoryError:
            warnings.warn(
                "fast_executemany execution failed.  Retrying with default executemany.  "
                "See https://github.com/BritishGeologicalSurvey/etlhelper/issues/86 for more information"
            )
            cursor.fast_executemany = False
            cursor.executemany(query, chunk)
        except TypeError:
            msg = ("pyodbc driver for MS SQL only supports positional placeholders.  "
                   "Use namedtuple, tuple or list (via row_factory setting for copy_rows).")
            raise ETLHelperInsertError(msg)



"""Integration tests for SQLite database.
These currently run against internal BGS instance.
"""
# pylint: disable=unused-argument, missing-docstring
import os
import sqlite3
import sys
from collections import namedtuple
from textwrap import dedent

import pytest

from etlhelper import (
    DbParams,
    connect,
    copy_rows,
    copy_table_rows,
    execute,
    fetchall,
    generate_insert_sql,
    load,
)
from etlhelper.utils import (
    table_info,
    Column,
)
from etlhelper.row_factories import namedtuple_row_factory
from etlhelper.exceptions import (
    ETLHelperConnectionError,
    ETLHelperInsertError,
    ETLHelperQueryError
)

# -- Tests here --


def test_connect(sqlitedb):
    conn = connect(sqlitedb)
    assert isinstance(conn, sqlite3.Connection)
    assert os.path.isfile(sqlitedb.filename)


@pytest.mark.skipif(sys.platform != 'linux', reason='Requires Linux OS')
def test_bad_connect(tmpdir):
    # Attempting to create file in non-existent directory should fail
    try:
        db_params = DbParams(dbtype='SQLITE', filename='/does/not/exist')
        with pytest.raises(ETLHelperConnectionError):
            connect(db_params)
    finally:
        # Restore permissions prior to cleanup
        os.chmod(tmpdir, 0o666)


def test_bad_select(testdb_conn):
    select_sql = "SELECT * FROM bad_table"
    with pytest.raises(ETLHelperQueryError):
        execute(select_sql, testdb_conn)


def test_bad_insert(testdb_conn):
    insert_sql = "INSERT INTO bad_table (id) VALUES (1)"
    with pytest.raises(ETLHelperQueryError):
        execute(insert_sql, testdb_conn)


def test_bad_constraint(test_tables, testdb_conn):
    # src already has a row with id=1
    insert_sql = "INSERT INTO src (id) VALUES (1)"
    with pytest.raises(ETLHelperQueryError):
        execute(insert_sql, testdb_conn)


def test_copy_rows_happy_path(test_tables, testdb_conn, test_table_data_dict):
    # Arrange and act
    select_sql = "SELECT * FROM src"
    insert_sql = INSERT_SQL.format(tablename='dest')
    copy_rows(select_sql, testdb_conn, insert_sql, testdb_conn, row_factory=namedtuple_row_factory)

    # Assert
    sql = "SELECT * FROM dest"
    result = fetchall(sql, testdb_conn)

    assert result == test_table_data_dict


def test_copy_table_rows_happy_path(test_tables, testdb_conn, test_table_data_dict):
    # Arrange and act
    copy_table_rows('src', testdb_conn, testdb_conn, target='dest')

    # Assert
    sql = "SELECT * FROM dest"
    result = fetchall(sql, testdb_conn)

    assert result == test_table_data_dict


def test_copy_table_rows_on_error(test_tables, testdb_conn, test_table_data_dict):
    # Arrange
    duplicate_id_row_sql = """
       INSERT INTO dest (id, value)
       VALUES (
         1,
         1.234
         )""".strip()
    execute(duplicate_id_row_sql, testdb_conn)

    # Act
    errors = []
    copy_table_rows('src', testdb_conn, testdb_conn, target='dest',
                    on_error=errors.extend)

    # Assert
    sql = "SELECT * FROM dest"
    result = fetchall(sql, testdb_conn)

    # Check that first row was caught as error
    row, exception = errors[0]
    assert row["id"] == 1
    assert "unique" in str(exception).lower()

    # Check that other rows were inserted correctly
    assert result[1:] == test_table_data_dict[1:]


def test_fetchall_with_parameters(test_tables, testdb_conn):
    # parameters=None is tested by default in other tests

    # Bind by index
    sql = "SELECT * FROM src where ID = ?"
    result = fetchall(sql, testdb_conn, parameters=(1,))
    assert len(result) == 1
    assert result[0]["id"] == 1


def test_copy_rows_bad_param_style(test_tables, testdb_conn, test_table_data_namedtuple):
    # Arrange and act
    select_sql = "SELECT * FROM src"
    insert_sql = BAD_PARAM_STYLE_SQL.format(tablename='dest')
    with pytest.raises(ETLHelperInsertError):
        copy_rows(select_sql, testdb_conn, insert_sql, testdb_conn)


def test_load_namedtuples(testdb_conn, test_tables, test_table_data_namedtuple):
    # Act
    load('dest', testdb_conn, test_table_data_namedtuple)

    # Assert
    sql = "SELECT * FROM dest"
    result = fetchall(sql, testdb_conn, row_factory=namedtuple_row_factory)

    assert result == test_table_data_namedtuple


def test_load_dicts(testdb_conn, test_tables, test_table_data_dict):
    # Act
    load('dest', testdb_conn, test_table_data_dict)

    # Assert
    sql = "SELECT * FROM dest"
    result = fetchall(sql, testdb_conn)

    assert result == test_table_data_dict


def test_generate_insert_sql_tuple(testdb_conn):
    # Act
    data = (1, 'one')
    with pytest.raises(ETLHelperInsertError,
                       match="Row is not a dictionary or namedtuple"):
        generate_insert_sql('my_table', data, testdb_conn)


def test_generate_insert_sql_namedtuple(testdb_conn):
    # Arrange
    TwoColumnRow = namedtuple('TwoColumnRow', ('id', 'data'))
    data = TwoColumnRow(1, 'one')
    expected = 'INSERT INTO my_table (id, data) VALUES (?, ?)'

    # Act
    sql = generate_insert_sql('my_table', data, testdb_conn)

    # Assert
    assert sql == expected


def test_generate_insert_sql_dictionary(testdb_conn):
    # Act
    data = {'id': 1, 'data': 'one'}

    expected = 'INSERT INTO my_table (id, data) VALUES (:id, :data)'

    # Act
    sql = generate_insert_sql('my_table', data, testdb_conn)

    # Assert
    assert sql == expected


def test_table_info_no_schema_no_duplicates(testdb_conn, test_tables):
    # Arrange
    expected = [
        Column(name='id', type='integer', not_null=0, has_default=0),
        Column(name='value', type='float', not_null=1, has_default=0),
        Column(name='simple_text', type='text', not_null=0, has_default=1),
        Column(name='utf8_text', type='text', not_null=0, has_default=0),
        Column(name='day', type='date', not_null=0, has_default=0),
        Column(name='date_time', type='timestamp', not_null=0, has_default=0)
    ]

    # Act
    columns = table_info('src', testdb_conn)

    # Assert
    assert columns == expected


def test_table_info_with_schema_no_duplicates(testdb_conn, test_tables):
    # Arrange
    expected = [
        Column(name='id', type='integer', not_null=0, has_default=0),
        Column(name='value', type='float', not_null=1, has_default=0),
        Column(name='simple_text', type='text', not_null=0, has_default=1),
        Column(name='utf8_text', type='text', not_null=0, has_default=0),
        Column(name='day', type='date', not_null=0, has_default=0),
        Column(name='date_time', type='timestamp', not_null=0, has_default=0)
    ]

    # Act
    columns = table_info('src', testdb_conn, schema='etlhelper')

    # Assert
    assert columns == expected


def test_table_info_bad_table_name_no_schema(testdb_conn, test_tables):
    # Arrange, act and assert
    with pytest.raises(ETLHelperQueryError, match=r"Table name 'bad_table' not found."):
        table_info('bad_table', testdb_conn)


def test_table_info_bad_table_name_with_schema(testdb_conn, test_tables):
    # Arrange, act and assert
    with pytest.raises(ETLHelperQueryError, match=r"Table name 'etlhelper.bad_table' not found."):
        table_info('bad_table', testdb_conn, schema='etlhelper')


# -- Fixtures here --

INSERT_SQL = dedent("""
    INSERT INTO {tablename} (id, value, simple_text, utf8_text,
      day, date_time)
    VALUES
      (?, ?, ?, ?, ?, ?)
      """).strip()

BAD_PARAM_STYLE_SQL = dedent("""
    INSERT INTO {tablename} (id, value, simple_text, utf8_text,
      day, date_time)
    VALUES
      (%s, %s, %s, %s, %s, %s)
      """).strip()


@pytest.fixture(scope='function')
def sqlitedb(tmp_path):
    """Get DbParams for temporary SQLite database."""
    filename = f'{tmp_path.absolute()}.db'
    yield DbParams(dbtype='SQLITE', filename=filename)


@pytest.fixture(scope='function')
def testdb_conn(sqlitedb):
    """Get connection to test SQLite database."""
    with connect(sqlitedb, detect_types=sqlite3.PARSE_DECLTYPES) as conn:
        # PARSE_DECLTYPES makes SQLite return TIMESTAMP columns as Python datetimes
        yield conn


@pytest.fixture(scope='function')
def test_tables(test_table_data_namedtuple, testdb_conn):
    """
    Create a table and fill with test data.  Teardown after the yield drops it
    again.
    """
    # Define SQL queries
    drop_src_sql = "DROP TABLE src"
    create_src_sql = dedent("""
        CREATE TABLE src
          (
            id integer primary key,
            value float not null,
            simple_text text default 'default',
            utf8_text text,
            day date,
            date_time timestamp
          )
          """).strip()
    drop_dest_sql = drop_src_sql.replace('src', 'dest')
    create_dest_sql = create_src_sql.replace('src', 'dest')

    # Create table and populate with test data
    # Unlike other databases, SQLite cursors cannot be used as context managers
    cursor = testdb_conn.cursor()
    # src table
    try:
        cursor.execute(drop_src_sql)
    except sqlite3.OperationalError:
        pass
    cursor.execute(create_src_sql)
    cursor.executemany(INSERT_SQL.format(tablename='src'),
                       test_table_data_namedtuple)
    # dest table
    try:
        cursor.execute(drop_dest_sql)
    except sqlite3.OperationalError:
        # Error if table doesn't exist
        pass
    cursor.execute(create_dest_sql)

    testdb_conn.commit()

    # Return control to calling function until end of test
    yield

    # Tear down the table after test completes
    cursor.execute(drop_src_sql)
    cursor.execute(drop_dest_sql)
    testdb_conn.commit()



"""Test for etl extract functions.  Extraction is done via the iter_chunks
function and those that call it.
These are run against PostgreSQL."""
# pylint: disable=unused-argument, missing-docstring
import datetime
import pytest
import time
from unittest.mock import (
    Mock,
    sentinel,
)

import etlhelper.etl as etlhelper_etl
from etlhelper import (
    iter_chunks,
    iter_rows,
    execute,
    fetchone,
    fetchall,
)
from etlhelper.etl import (
    ETLHelperExtractError,
    ETLHelperQueryError,
)
from etlhelper.row_factories import (
    dict_row_factory,
    list_row_factory,
    namedtuple_row_factory,
    tuple_row_factory,
)


@pytest.mark.parametrize(
    ["chunk_size", "slices"],
    [
        (1, [slice(0, 1), slice(1, 2), slice(2, 3)]),
        (2, [slice(0, 2), slice(2, 3)]),
        (5000, [slice(0, 3)]),
    ],
)
def test_iter_chunks(
    pgtestdb_test_tables,
    pgtestdb_conn,
    test_table_data_namedtuple,
    chunk_size,
    slices,
):
    # Arrange
    sql = "SELECT * FROM src"

    # Act
    result = [
        list(chunk) for chunk in
        iter_chunks(sql, pgtestdb_conn, chunk_size=chunk_size, row_factory=namedtuple_row_factory)
    ]

    # Assert
    expected = [test_table_data_namedtuple[s] for s in slices]
    assert result == expected


def test_iter_rows_happy_path(
    pgtestdb_test_tables,
    pgtestdb_conn,
    test_table_data_namedtuple,
):
    sql = "SELECT * FROM src"
    result = iter_rows(sql, pgtestdb_conn, row_factory=namedtuple_row_factory)
    assert list(result) == test_table_data_namedtuple


def test_iter_rows_with_parameters(
    pgtestdb_test_tables,
    pgtestdb_conn,
    test_table_data_namedtuple,
):
    # parameters=None is tested by default in other tests

    # Bind by index
    sql = "SELECT * FROM src where ID = %s"
    result = iter_rows(sql, pgtestdb_conn, parameters=(1,), row_factory=namedtuple_row_factory)
    assert list(result) == [test_table_data_namedtuple[0]]

    # Bind by name
    sql = "SELECT * FROM src where ID = %(identifier)s"
    result = iter_rows(sql, pgtestdb_conn, parameters={"identifier": 1}, row_factory=namedtuple_row_factory)
    assert list(result) == [test_table_data_namedtuple[0]]


def test_iter_rows_transform(
    pgtestdb_test_tables,
    pgtestdb_conn,
    test_table_data_namedtuple,
):
    # Arrange
    sql = "SELECT * FROM src"

    def my_transform(rows):
        # Simple transform function that changes size and number of rows
        return [row.id for row in rows if row.id > 1]

    # Act
    result = iter_rows(sql, pgtestdb_conn, transform=my_transform, row_factory=namedtuple_row_factory)

    # Assert
    expected = [row[0] for row in test_table_data_namedtuple if row[0] > 1]
    assert list(result) == expected


def test_iter_rows_namedtuple_factory(
    pgtestdb_test_tables,
    pgtestdb_conn,
    test_table_data_namedtuple,
):
    sql = "SELECT * FROM src"
    result = iter_rows(sql, pgtestdb_conn, row_factory=namedtuple_row_factory)
    row = list(result)[0]

    assert row.id == 1
    assert row.value == 1.234
    assert row.simple_text == "text"
    assert row.utf8_text == "Öæ°\nz"
    assert row.day == datetime.date(2018, 12, 7)

    # The final assertion is skipped because the test fails, even though the
    # correct value has been assigned.  I don't know why.
    # assert row.date_time == datetime.datetime(2018, 12, 9, 13, 1, 59)


@pytest.mark.parametrize(
    ["row_factory", "expected"],
    [
        (dict_row_factory, {
            "id": 1, "value": 1.234, "simple_text": "text",
            "utf8_text": "Öæ°\nz", "day": datetime.date(2018, 12, 7),
            "date_time": datetime.datetime(2018, 12, 7, 13, 1, 59)}),
        (tuple_row_factory, (
            1, 1.234, "text", "Öæ°\nz", datetime.date(2018, 12, 7),
            datetime.datetime(2018, 12, 7, 13, 1, 59))),
        (list_row_factory, [
            1, 1.234, "text", "Öæ°\nz", datetime.date(2018, 12, 7),
            datetime.datetime(2018, 12, 7, 13, 1, 59)]),
    ],
)
def test_iter_rows_other_row_factories(
    row_factory, expected,
    pgtestdb_test_tables,
    pgtestdb_conn,
):
    sql = "SELECT * FROM src LIMIT 1"
    result = iter_rows(sql, pgtestdb_conn, row_factory=row_factory)

    assert list(result) == [expected]


def test_iter_rows_no_results(pgtestdb_test_tables, pgtestdb_conn):
    sql = "SELECT * FROM src WHERE id = -1"
    result = iter_rows(sql, pgtestdb_conn)
    assert list(result) == []


def test_iter_rows_bad_query(pgtestdb_test_tables, pgtestdb_conn):
    sql = "SELECT * FROM public.this_does_not_exist"
    with pytest.raises(ETLHelperExtractError):
        result = iter_rows(sql, pgtestdb_conn)
        list(result)  # Call list to activate returned generator


def test_fetchone_happy_path(
    pgtestdb_test_tables,
    pgtestdb_conn,
    test_table_data_dict,
):
    sql = "SELECT * FROM src"
    result = fetchone(sql, pgtestdb_conn)
    assert result == test_table_data_dict[0]


def test_fetchone_none(
    pgtestdb_test_tables,
    pgtestdb_conn,
    test_table_data_namedtuple,
):
    sql = "SELECT * FROM src WHERE id=999"
    result = fetchone(sql, pgtestdb_conn)
    assert result is None


def test_fetchone_closes_transaction(pgtestdb_conn):
    sql = "SELECT now() AS time"

    # Call the time function
    time1 = fetchone(sql, pgtestdb_conn)

    # Call it again after a delay
    time.sleep(0.001)
    time2 = fetchone(sql, pgtestdb_conn)

    # If the connection was not closed and the same transaction was used
    # then the times would be the same
    assert time2["time"] > time1["time"]


@pytest.mark.parametrize(
    "fetch_func",
    [fetchall],
)
def test_fetch_funcs_close_transaction(pgtestdb_conn, fetch_func):
    sql = "SELECT now() AS time"

    # Call the time function
    time1 = fetch_func(sql, pgtestdb_conn)[0]

    # Call it again after a delay
    time.sleep(0.001)
    time2 = fetch_func(sql, pgtestdb_conn)[0]

    # If the connection was not closed and the same transaction was used
    # then the times would be the same
    assert time2["time"] > time1["time"]


def test_fetchall_happy_path(
    pgtestdb_test_tables,
    pgtestdb_conn,
    test_table_data_dict,
):
    sql = "SELECT * FROM src"
    result = fetchall(sql, pgtestdb_conn)
    assert result == test_table_data_dict


@pytest.mark.parametrize(
    "fetchmethod",
    ["fetchone", "fetchall"],
)
def test_arguments_passed_to_iter_rows(
    monkeypatch,
    fetchmethod,
    pgtestdb_test_tables,
    pgtestdb_conn,
):
    """Each of these functions calls iter_rows.  This tests check that the
    correct parameters are passed through, which confirms parameters,
    row_factory and transform can all be used."""

    # Arrange - patch 'iter_rows' function within etlhelper.etl module
    mock_iter_rows = Mock()
    monkeypatch.setattr(etlhelper_etl, "iter_rows", mock_iter_rows)

    # Sentinel items are unique so confirm object that was passed through
    sql = sentinel.sql
    parameters = sentinel.parameters
    transform = sentinel.transform
    chunk_size = sentinel.chunk_size

    # Act
    # getattr returns fetchmethod, which we call with given parameters
    # Use real connection parameters to ensure we reach the call to iter_rows
    # Use dict_row_factory to demonstrate the default (namedtuple_row_factory)
    # isn't called
    try:
        getattr(etlhelper_etl, fetchmethod)(
            sql,
            pgtestdb_conn,
            parameters=parameters,
            row_factory=dict_row_factory,
            transform=transform,
            chunk_size=chunk_size,
        )
    except TypeError:
        # We expect an error here as the mock_iter_rows breaks the functions
        # that called it.
        pass

    # Assert
    mock_iter_rows.assert_called_once_with(
        sql,
        pgtestdb_conn,
        parameters=parameters,
        row_factory=dict_row_factory,
        transform=transform,
        chunk_size=chunk_size,
    )


def test_execute_happy_path(pgtestdb_test_tables, pgtestdb_conn):
    # Arrange
    sql = "DELETE FROM src;"

    # Act
    execute(sql, pgtestdb_conn)

    # Assert
    result = fetchall("SELECT * FROM src;", pgtestdb_conn)
    assert result == []


def test_execute_with_params(
    pgtestdb_test_tables,
    pgtestdb_conn,
    test_table_data_dict,
):
    # Arrange
    sql = "DELETE FROM src WHERE id = %s;"
    params = [1]
    expected = test_table_data_dict[1:]

    # Act
    execute(sql, pgtestdb_conn, parameters=params)

    # Assert
    result = fetchall("SELECT * FROM src;", pgtestdb_conn)
    assert result == expected


def test_execute_bad_query(pgtestdb_test_tables, pgtestdb_conn):
    sql = "DELETE * FROM this_does_not_exist"
    with pytest.raises(ETLHelperQueryError):
        execute(sql, pgtestdb_conn)



"""
Row factories are functions that process row data as it comes from the database.
These are applied by the iter_rows function.

A row_factory function must:
  + accept a cursor object as an input
  + only use methods on the cursor that are described by DBAPI
  + return a function that takes a tuple

"""
import re
from collections import namedtuple
from warnings import warn


def namedtuple_row_factory(cursor):
    """
    Return function to convert output row to a named tuple.

    Named tuples allow access to attributes via both position (e.g. row[0]) or
    name (using dot notation, e.g. row.id).  They are immutable, so cannot be
    modified directly in transform functions.  Insert statements based on named
    tuples must use positional placeholders for parameters (e.g. ?, :1, %s).
    """
    column_names = [d[0] for d in cursor.description]

    try:
        Row = namedtuple('Row', field_names=column_names)
    except ValueError:
        Row = namedtuple('Row', field_names=column_names, rename=True)
        renamed_columns = _find_renamed_columns(Row, column_names)
        warn("One or more columns have been renamed. Names that cannot be "
             "converted to namedtuple attributes are replaced by indices. To "
             "prevent column renaming, either provide alias in SQL query, "
             "e.g. 'SELECT count(*) AS c', or use dict_row_factory. ")
        warn(f"{renamed_columns}")

    def create_row(row):
        return Row(*row)

    return create_row


def dict_row_factory(cursor):
    """
    Return function to convert output row to a dictionary.

    Dictionaries allow access to attributes via name (using key notation, e.g.
    row["id"].  They are mutable, so are convenient to modify directly in
    transform functions.  Insert statements based on dictionaries must use
    named placeholders for parameters (e.g. :id, %(id)s).
    """
    column_names = [d[0] for d in cursor.description]

    def create_row(row):
        row_dict = {}
        for i, column_name in enumerate(column_names):
            row_dict[column_name] = row[i]
        return row_dict

    return create_row


def tuple_row_factory(cursor):
    """
    Return function to convert output row to a tuple.

    Tuples allow access to attributes via position (e.g. row[0]).  They are
    immutable, so cannot be modified directly in transform functions.  Insert
    statements based on tuples must use positional placeholders for parameters
    (e.g. ?, :1, %s).

    As the DBAPI default is already to return rows as tuples, using the tuple
    row factory minimises processing overhead.
    """
    def create_row(row):
        return row

    return create_row


def list_row_factory(cursor):
    """
    Return function to convert output row to a list.

    Lists allow access to attributes via position (e.g. row[0]).  They are
    mutable, so are convient to modify directly in transform functions.  Insert
    statements based on lists must use positional placeholders for parameters
    (e.g. ?, :1, %s).
    """
    def create_row(row):
        return list(row)

    return create_row


def _find_renamed_columns(row_class, column_names):
    regex = re.compile(r'^_\d+$')

    renamed_column_ids = [int(f.replace("_", "")) for f in row_class._fields if regex.match(f)]

    result = '\n'.join(f'{column_names[idx]} was renamed to {row_class._fields[idx]}'
                       for idx in renamed_column_ids)

    return result



"""ETL Helper script to demonstrate table_info."""
from etlhelper.utils import table_info
from my_databases import ORACLEDB

with ORACLEDB.connect("ORA_PASSWORD") as conn:
    columns = table_info("my_table", conn, schema="my_schema")



"""DB Helper factory

Factory pattern that generates a DbHelper for each DB type

"""
from functools import lru_cache

from etlhelper.db_helpers.oracle import OracleDbHelper
from etlhelper.db_helpers.postgres import PostgresDbHelper
from etlhelper.db_helpers.mssql import MSSQLDbHelper
from etlhelper.db_helpers.sqlite import SQLiteDbHelper
from etlhelper.exceptions import ETLHelperHelperError


class DbHelperFactory():
    """
    The DB Helper Factory class.
    """
    def __init__(self):
        self.helpers = {}
        self._conn_types: dict[str, str] = {}

    def register_helper(self, dbtype, conn_type, db_helper):
        """
        Store db helper in internal list.
        """
        self.helpers[dbtype] = db_helper
        self._conn_types[conn_type] = dbtype

    def from_db_params(self, db_params):
        """
        Return initialised db_helper
        """
        if not hasattr(db_params, 'dbtype'):
            msg = f"Expected DbParams-like object, got {type(db_params)}"
            raise ETLHelperHelperError(msg)

        return self.from_dbtype(db_params.dbtype)

    def from_conn(self, conn):
        """
        Return initialised db_helper based on connection.
        """
        if not hasattr(conn, 'cursor'):
            msg = f"Expected connection-like object, got {type(conn)}"
            raise ETLHelperHelperError(msg)

        try:
            conn_type = str(conn.__class__)
            dbtype = self._conn_types[conn_type]
        except KeyError:
            msg = f"Unsupported connection type: {conn_type}"
            raise ETLHelperHelperError(msg)
        return self.from_dbtype(dbtype)

    @lru_cache(maxsize=16)
    def from_dbtype(self, dbtype):
        """
        Return initialised db helper based on type
        """
        try:
            helper = self.helpers[dbtype]()
        except KeyError:
            msg = f"Unsupported DbParams.dbtype: {dbtype}"
            raise ETLHelperHelperError(msg)
        return helper


DB_HELPER_FACTORY = DbHelperFactory()
DB_HELPER_FACTORY.register_helper('ORACLE', "<class 'oracledb.Connection'>",
                                  OracleDbHelper)
DB_HELPER_FACTORY.register_helper('PG', "<class 'psycopg2.extensions.connection'>",
                                  PostgresDbHelper)
DB_HELPER_FACTORY.register_helper('MSSQL', "<class 'pyodbc.Connection'>",
                                  MSSQLDbHelper)
DB_HELPER_FACTORY.register_helper('SQLITE', "<class 'sqlite3.Connection'>",
                                  SQLiteDbHelper)



"""
Tests for implementation of the abort() function in both extracting and
loading data.
"""
import logging
import sqlite3
from concurrent.futures import ThreadPoolExecutor
from time import sleep

import pytest

from etlhelper import (
    abort_etlhelper_threads,
    execute,
    executemany,
    fetchall,
)
from etlhelper.exceptions import ETLHelperAbort

logger = logging.getLogger('abort_test')
logger.setLevel(logging.INFO)


def do_fetchall_etl(db):
    # Example of a fetchall ETL job

    def log_chunks_slowly(chunk):
        logger.info("Processing chunk")
        # Add a delay to result fetching to give time for abort call
        sleep(0.05)
        return chunk

    with sqlite3.connect(db) as conn:
        result = fetchall("SELECT * FROM test", conn,
                          transform=log_chunks_slowly, chunk_size=1)

    return result


def do_executemany_etl(db):
    # Example of an executemany ETL job

    # Create generator to produce rows with delay
    def slow_row_generator(rows_generated):
        for i in range(10):
            sleep(0.05)
            logger.info("Generating row")
            row = (i,)
            rows_generated.append(row)
            yield row

    rows_generated = []
    with sqlite3.connect(db) as conn:
        executemany("INSERT INTO test VALUES (?)", conn,
                    slow_row_generator(rows_generated), chunk_size=1)

    return rows_generated


@pytest.mark.parametrize("do_etl, log_keyword", [
    (do_fetchall_etl, 'iter_chunks'),
    (do_executemany_etl, 'executemany')
    ])
def test_abort_etlhelper_threads(do_etl, log_keyword, tmpdir, caplog):
    """
    Transfer one row at a time to/from temporary database within a thread pool,
    call abort() then assert that exception was raised and not all rows were
    processed.
    """
    # Arrange
    # Create and populate a temporary SQLite database
    db = str(tmpdir / 'test.db')  # sqlite3 on Python 3.6 requires a string

    with sqlite3.connect(db) as conn:
        execute("CREATE TABLE test (id INTEGER)", conn)
        rows = ((row,) for row in range(10))
        executemany("INSERT INTO test VALUES (?)", conn, rows=rows)

    # Act and assert
    # Do the ETL within a thread confirm it was aborted
    with ThreadPoolExecutor(max_workers=1) as executor:
        future = executor.submit(do_etl, db)

        # Call abort after short delay
        sleep(0.2)
        abort_etlhelper_threads()

        # Exceptions from threads are raised when result is retrieved
        with pytest.raises(ETLHelperAbort, match=log_keyword):
            _ = future.result()

    # The number of log records is a proxy for the rows handled
    assert len(caplog.records) < 10

    # Redo the ETL without abort to confirm that abort event was cleared
    with ThreadPoolExecutor(max_workers=1) as executor:
        future = executor.submit(do_etl, db)
        result = future.result()

    assert len(result) == 10



"""Integration tests for Oracle database.
These currently run against internal BGS instance.
"""
# pylint: disable=unused-argument, missing-docstring
import os
from collections import namedtuple
from textwrap import dedent

import oracledb
import pytest

from etlhelper import (
    DbParams,
    connect,
    copy_rows,
    copy_table_rows,
    execute,
    fetchall,
    generate_insert_sql,
    load,
)
from etlhelper.utils import (
    table_info,
    Column,
)
from etlhelper.row_factories import namedtuple_row_factory
from etlhelper.exceptions import (
    ETLHelperConnectionError,
    ETLHelperInsertError,
    ETLHelperQueryError
)

# Skip these tests if database is unreachable
if not os.getenv('TEST_ORACLE_HOST'):
    pytest.skip('Oracle test database is not defined', allow_module_level=True)

ORADB = DbParams.from_environment(prefix='TEST_ORACLE_')
if not ORADB.is_reachable():
    pytest.skip('Oracle test database is unreachable', allow_module_level=True)


# -- Tests here --

def test_connect():
    conn = connect(ORADB, 'TEST_ORACLE_PASSWORD')
    assert isinstance(conn, oracledb.Connection)


def test_connect_wrong_password(monkeypatch):
    monkeypatch.setitem(os.environ, 'TEST_ORACLE_PASSWORD', 'bad_password')
    with pytest.raises(ETLHelperConnectionError):
        connect(ORADB, 'TEST_ORACLE_PASSWORD')


def test_bad_select(testdb_conn):
    select_sql = "SELECT * FROM bad_table"
    with pytest.raises(ETLHelperQueryError):
        execute(select_sql, testdb_conn)


def test_bad_insert(testdb_conn):
    insert_sql = "INSERT INTO bad_table (id) VALUES (1)"
    with pytest.raises(ETLHelperQueryError):
        execute(insert_sql, testdb_conn)


def test_bad_constraint(test_tables, testdb_conn):
    # src already has a row with id=1
    insert_sql = "INSERT INTO src (id) VALUES (1)"
    with pytest.raises(ETLHelperQueryError):
        execute(insert_sql, testdb_conn)


def test_copy_rows_happy_path(test_tables, testdb_conn, test_table_data_dict):
    # Arrange and act
    select_sql = "SELECT * FROM src"
    insert_sql = INSERT_SQL.format(tablename='dest')
    copy_rows(select_sql, testdb_conn, insert_sql, testdb_conn)

    # Assert
    sql = "SELECT * FROM dest"
    result = fetchall(sql, testdb_conn)

    # Fix result date and datetime strings to native classes
    fixed_dates = []
    for row in result:
        row = {key.lower(): value for key, value in row.items()}
        row["day"] = row["day"].date()
        fixed_dates.append(row)

    assert fixed_dates == test_table_data_dict


def test_copy_table_rows_happy_path(test_tables, testdb_conn, test_table_data_dict):
    # Arrange and act
    copy_table_rows('src', testdb_conn, testdb_conn, target='dest')

    # Assert
    sql = "SELECT * FROM dest"
    result = fetchall(sql, testdb_conn)

    # Fix result date and datetime strings to native classes
    fixed_dates = []
    for row in result:
        row = {key.lower(): value for key, value in row.items()}
        row["day"] = row["day"].date()
        fixed_dates.append(row)

    assert fixed_dates == test_table_data_dict


def test_copy_table_rows_on_error(test_tables, testdb_conn, test_table_data_dict):
    # Arrange
    duplicate_id_row_sql = """
       INSERT INTO dest (id, value)
       VALUES (
         1,
         1.234
         )""".strip()
    execute(duplicate_id_row_sql, testdb_conn)

    # Act
    errors = []
    copy_table_rows('src', testdb_conn, testdb_conn, target='dest',
                    on_error=errors.extend)

    # Assert
    sql = "SELECT * FROM dest"
    result = fetchall(sql, testdb_conn)

    # Check that first row was caught as error, noting that Oracle
    # changes the case of column names
    row, exception = errors[0]
    assert row["ID"] == 1
    assert "unique" in str(exception).lower()

    # Check that other rows were inserted correctly
    # Fix result date and datetime strings to native classes
    fixed_dates = []
    for row in result[1:]:
        row = {key.lower(): value for key, value in row.items()}
        row["day"] = row["day"].date()
        fixed_dates.append(row)

    assert fixed_dates == test_table_data_dict[1:]


def test_fetchall_with_parameters(test_tables, testdb_conn):
    # parameters=None is tested by default in other tests

    # Bind by index
    sql = "SELECT * FROM src where ID = :1"
    result = fetchall(sql, testdb_conn, parameters=(1,))
    assert len(result) == 1
    assert result[0]["ID"] == 1


def test_copy_rows_bad_param_style(test_tables, testdb_conn):
    # Arrange and act
    select_sql = "SELECT * FROM src"
    insert_sql = BAD_PARAM_STYLE_SQL.format(tablename='dest')
    with pytest.raises(ETLHelperInsertError):
        copy_rows(select_sql, testdb_conn, insert_sql, testdb_conn)


def test_load_namedtuples(testdb_conn, test_tables, test_table_data_namedtuple):
    # Arrange
    # Convert to plain tuples as ORACLE makes column names upper case
    expected = [tuple(row) for row in test_table_data_namedtuple]

    # Act
    load('dest', testdb_conn, test_table_data_namedtuple)

    # Assert
    sql = "SELECT * FROM dest"
    result = fetchall(sql, testdb_conn, row_factory=namedtuple_row_factory)

    # Fix result date and datetime strings to native classes
    fixed_dates = []
    for row in result:
        fixed_dates.append((
            *row[:4],
            row.DAY.date(),
            row.DATE_TIME
        ))

    assert fixed_dates == expected


def test_load_dicts(testdb_conn, test_tables, test_table_data_dict):
    # Act
    load('dest', testdb_conn, test_table_data_dict)

    # Assert
    sql = "SELECT * FROM dest"
    result = fetchall(sql, testdb_conn)

    # Fix result date and datetime strings to native classes
    fixed_dates = []
    for row in result:
        row = {key.lower(): value for key, value in row.items()}
        row["day"] = row["day"].date()
        fixed_dates.append(row)

    assert fixed_dates == test_table_data_dict


def test_generate_insert_sql_tuple(testdb_conn):
    # Act
    data = (1, 'one')
    with pytest.raises(ETLHelperInsertError,
                       match="Row is not a dictionary or namedtuple"):
        generate_insert_sql('my_table', data, testdb_conn)


def test_generate_insert_sql_namedtuple(testdb_conn):
    # Arrange
    TwoColumnRow = namedtuple('TwoColumnRow', ('id', 'data'))
    data = TwoColumnRow(1, 'one')
    expected = 'INSERT INTO my_table (id, data) VALUES (:1, :2)'

    # Act
    sql = generate_insert_sql('my_table', data, testdb_conn)

    # Assert
    assert sql == expected


def test_generate_insert_sql_dictionary(testdb_conn):
    # Act
    data = {'id': 1, 'data': 'one'}
    expected = 'INSERT INTO my_table (id, data) VALUES (:id, :data)'

    # Act
    sql = generate_insert_sql('my_table', data, testdb_conn)

    # Assert
    assert sql == expected


def test_table_info_no_schema_no_duplicates(testdb_conn, test_tables):
    # Arrange
    expected = [
        Column(name='ID', type='NUMBER', not_null=0, has_default=0),
        Column(name='VALUE', type='NUMBER', not_null=1, has_default=0),
        Column(name='SIMPLE_TEXT', type='VARCHAR2', not_null=0, has_default=1),
        Column(name='UTF8_TEXT', type='VARCHAR2', not_null=0, has_default=0),
        Column(name='DAY', type='DATE', not_null=0, has_default=0),
        Column(name='DATE_TIME', type='DATE', not_null=0, has_default=0)
    ]

    # Act
    columns = table_info('src', testdb_conn)

    # Assert
    assert columns == expected


def test_table_info_with_schema_no_duplicates(testdb_conn, test_tables):
    # Arrange
    expected = [
        Column(name='ID', type='NUMBER', not_null=0, has_default=0),
        Column(name='VALUE', type='NUMBER', not_null=1, has_default=0),
        Column(name='SIMPLE_TEXT', type='VARCHAR2', not_null=0, has_default=1),
        Column(name='UTF8_TEXT', type='VARCHAR2', not_null=0, has_default=0),
        Column(name='DAY', type='DATE', not_null=0, has_default=0),
        Column(name='DATE_TIME', type='DATE', not_null=0, has_default=0)
    ]

    # Act
    columns = table_info('src', testdb_conn, schema='etlhelper')

    # Assert
    assert columns == expected


def test_table_info_bad_table_name_no_schema(testdb_conn, test_tables):
    # Arrange, act and assert
    with pytest.raises(ETLHelperQueryError, match=r"Table name 'bad_table' not found."):
        table_info('bad_table', testdb_conn)


def test_table_info_bad_table_name_with_schema(testdb_conn, test_tables):
    # Arrange, act and assert
    with pytest.raises(ETLHelperQueryError, match=r"Table name 'etlhelper.bad_table' not found."):
        table_info('bad_table', testdb_conn, schema='etlhelper')


@pytest.mark.parametrize('fetch_lobs', (
        True,  # default for oracledb
        False  # default for etlhelper
    ))
def test_lob_io(testdb_conn, test_blob_table, fetch_lobs):
    """
    These tests confirm that:
      + Changing the fetch_lobs parameter before creating a cursor affects
        the returned data type for the subsequent query
      + It is not necessary to change the setting prior to connecting
      + CLOBs longer than 4000 characters are returned in full
    """

    insert_sql = dedent("""
        INSERT INTO blob_src (id, my_clob, my_blob)
        VALUES (:1, :2, :3)
        """).strip()
    big_text = "abcd" * 10000
    big_bytes = big_text.encode('utf8')

    oracledb.defaults.fetch_lobs = fetch_lobs

    # populate table with data
    with testdb_conn.cursor() as cursor:
        cursor.execute(insert_sql, (1, big_text, big_bytes))
    testdb_conn.commit()

    # fetch data
    with testdb_conn.cursor() as cursor:
        cursor.execute("SELECT my_clob, my_blob FROM blob_src")
        big_text_returned, big_bytes_returned = cursor.fetchone()

        if fetch_lobs:
            assert isinstance(big_text_returned, oracledb.LOB)
            assert isinstance(big_bytes_returned, oracledb.LOB)
        else:
            assert isinstance(big_text_returned, str)
            assert big_text_returned == big_text
            assert isinstance(big_bytes_returned, bytes)
            assert big_bytes_returned == big_bytes


# -- Fixtures here --

INSERT_SQL = dedent("""
    INSERT INTO {tablename} (id, value, simple_text, utf8_text,
      day, date_time)
    VALUES
      (:id, :value, :simple_text, :utf8_text, :day, :date_time)
      """).strip()

BAD_PARAM_STYLE_SQL = dedent("""
    INSERT INTO {tablename} (id, value, simple_text, utf8_text,
      day, date_time)
    VALUES
      (?, ?, ?, ?, ?, ?)
      """).strip()


@pytest.fixture(scope='function')
def testdb_conn():
    """Get connection to test Oracle database."""
    with connect(ORADB, 'TEST_ORACLE_PASSWORD') as conn:
        yield conn


@pytest.fixture(scope='function')
def test_tables(test_table_data_namedtuple, testdb_conn):
    """
    Create a table and fill with test data.  Teardown after the yield drops it
    again.
    """
    # Define SQL queries
    drop_src_sql = "DROP TABLE src"
    create_src_sql = dedent("""
        CREATE TABLE src
          (
            id NUMBER UNIQUE,
            value NUMBER not null,
            simple_text VARCHAR2(100) default 'default',
            utf8_text VARCHAR2(100),
            day DATE,
            date_time DATE
          )
          """).strip()
    drop_dest_sql = drop_src_sql.replace('src', 'dest')
    create_dest_sql = create_src_sql.replace('src', 'dest')

    # Create table and populate with test data
    with testdb_conn.cursor() as cursor:
        # src table
        try:
            cursor.execute(drop_src_sql)
        except oracledb.DatabaseError:
            pass
        cursor.execute(create_src_sql)
        cursor.executemany(INSERT_SQL.format(tablename='src'),
                           test_table_data_namedtuple)
        # dest table
        try:
            cursor.execute(drop_dest_sql)
        except oracledb.DatabaseError:
            # Error if table doesn't exist
            pass
        cursor.execute(create_dest_sql)
    testdb_conn.commit()

    # Return control to calling function until end of test
    yield

    # Tear down the table after test completes
    with testdb_conn.cursor() as cursor:
        cursor.execute(drop_src_sql)
        cursor.execute(drop_dest_sql)
    testdb_conn.commit()


@pytest.fixture(scope='function')
def test_blob_table(testdb_conn):
    """
    Create a blob table and fill with test data.  Teardown after the yield drops it
    again.
    """
    # Define SQL queries
    create_src_sql = dedent("""
        CREATE TABLE blob_src
          (
            id NUMBER UNIQUE,
            my_clob CLOB,
            my_blob BLOB
          )
          """).strip()
    drop_src_sql = "DROP TABLE blob_src"

    # Create table
    with testdb_conn.cursor() as cursor:
        # src table
        try:
            cursor.execute(drop_src_sql)
        except oracledb.DatabaseError:
            pass
        cursor.execute(create_src_sql)
    testdb_conn.commit()

    # Return control to calling function until end of test
    yield

    # Tear down the table after test completes
    with testdb_conn.cursor() as cursor:
        cursor.execute(drop_src_sql)
    testdb_conn.commit()



"""Tests for etl logging.  These are run against PostgreSQL."""
# pylint: disable=unused-argument, missing-docstring
import logging
import pytest
import re

from etlhelper import (
    copy_rows,
    execute,
)
from etlhelper.row_factories import namedtuple_row_factory

NO_OUTPUT = []
INFO = [
    'Executing many (chunk_size=1)',
    'Fetching rows (chunk_size=1)',
    '1 rows processed (0 failed)',
    '2 rows processed (0 failed)',
    '3 rows processed (0 failed)',
    '3 rows returned',
    '3 rows processed in total']
INFO_AND_DEBUG = [
    'Executing many (chunk_size=1)',
    'Executing:\n'
    '\n'
    'INSERT INTO dest (id, value, simple_text, utf8_text,\n'
    '  day, date_time)\n'
    'VALUES\n'
    '  (%s, %s, %s, %s, %s, %s)\n'
    '  ;\n'
    '\n'
    'against:\n'
    '\n'
    "<connection object at ???; dsn: 'user=etlhelper_user password=xxx "
    "dbname=etlhelper host=??? port=???', closed: 0>",
    'Fetching rows (chunk_size=1)',
    'Fetching:\n'
    '\n'
    'SELECT * FROM src\n'
    '\n'
    'with parameters:\n'
    '\n'
    '()\n'
    '\n'
    'against:\n'
    '\n'
    "<connection object at ???; dsn: 'user=etlhelper_user password=xxx "
    "dbname=etlhelper host=??? port=???', closed: 0>",
    "First row: Row(id=1, value=1.234, simple_text='text', utf8_text='Öæ°\\nz', "
    'day=datetime.date(2018, 12, 7), date_time=datetime.datetime(2018, 12, 7, 13, '
    '1, 59))',
    '1 rows processed (0 failed)',
    '2 rows processed (0 failed)',
    '3 rows processed (0 failed)',
    '3 rows returned',
    '3 rows processed in total']


@pytest.mark.parametrize('level, expected', [
    (logging.DEBUG, INFO_AND_DEBUG),
    (logging.INFO, INFO),
    (logging.WARNING, NO_OUTPUT),
])
def test_logging_copy_rows(caplog, level, expected,
                           pgtestdb_conn, pgtestdb_test_tables,
                           pgtestdb_insert_sql, logger):
    # Arrange
    caplog.set_level(level, logger=logger.name)
    select_sql = "SELECT * FROM src"
    insert_sql = pgtestdb_insert_sql.replace('src', 'dest')

    # Act
    copy_rows(select_sql, pgtestdb_conn, insert_sql, pgtestdb_conn,
              chunk_size=1, row_factory=namedtuple_row_factory)

    # ID for connection object and hostname vary between tests
    # and test environments
    messages = [re.sub(r'object at .*;', 'object at ???;', m)
                for m in caplog.messages]
    messages = [re.sub(r'host=.*? ', 'host=??? ', m)
                for m in messages]
    messages = [re.sub(r'port=[0-9]{1,5}\'', 'port=???\'', m)
                for m in messages]

    # Assert
    for i, message in enumerate(messages):
        assert message == expected[i]


INFO_EXECUTE = ['Executing query']
INFO_AND_DEBUG_EXECUTE = [
    'Executing query',
    'Executing:\n'
    '\n'
    'SELECT 1 AS result;\n'
    '\n'
    'with parameters:\n'
    '\n'
    '()\n'
    '\n'
    'against:\n'
    '\n'
    "<connection object at ???; dsn: 'user=etlhelper_user password=xxx "
    "dbname=etlhelper host=??? port=???', closed: 0>"]


@pytest.mark.parametrize('level, expected', [
    (logging.DEBUG, INFO_AND_DEBUG_EXECUTE),
    (logging.INFO, INFO_EXECUTE),
    (logging.WARNING, NO_OUTPUT),
])
def test_logging_execute(caplog, level, expected, pgtestdb_conn, logger):
    # Arrange
    caplog.set_level(level, logger=logger.name)
    select_sql = "SELECT 1 AS result;"

    # Act
    execute(select_sql, pgtestdb_conn)
    # ID for connection object and hostname vary between tests
    # and test environments
    messages = [re.sub(r'object at .*;', 'object at ???;', m)
                for m in caplog.messages]
    messages = [re.sub(r'host=.*? ', 'host=??? ', m)
                for m in messages]
    messages = [re.sub(r'port=[0-9]{1,5}\'', 'port=???\'', m)
                for m in messages]

    # Assert
    for i, message in enumerate(messages):
        assert message == expected[i]



"""Unit tests for db_helpers module."""
import builtins
import sqlite3
from unittest.mock import Mock

import oracledb
import psycopg2
import pyodbc
import pytest

from etlhelper import DbParams
from etlhelper.db_helper_factory import DB_HELPER_FACTORY
from etlhelper.db_helpers import (
    MSSQLDbHelper,
    OracleDbHelper,
    PostgresDbHelper,
    SQLiteDbHelper
)
from etlhelper.exceptions import ETLHelperConnectionError

# pylint: disable=missing-docstring

ORACLEDB = DbParams(dbtype='ORACLE', host='server', port='1521',
                    dbname='testdb', user='testuser')

MSSQLDB = DbParams(dbtype='MSSQL', host='server', port='1521', dbname='testdb',
                   user='testuser', odbc_driver='test driver')

POSTGRESDB = DbParams(dbtype='PG', host='server', port='1521', dbname='testdb',
                      user='testuser')

SQLITEDB = DbParams(dbtype='SQLITE', filename='/myfile.db')


@pytest.mark.parametrize('helper, expected', [
    (OracleDbHelper, (oracledb.DatabaseError, oracledb.InterfaceError)),
    (MSSQLDbHelper, (pyodbc.DatabaseError, pyodbc.InterfaceError)),
    (PostgresDbHelper, (psycopg2.DatabaseError, psycopg2.InterfaceError)),
    (SQLiteDbHelper, (sqlite3.DatabaseError, sqlite3.InterfaceError))
])
def test_sql_exceptions(helper, expected):
    # DBAPI 2.0 specifies that all exceptions must inherit from DatabaseError
    # or InterfaceError. See https://peps.python.org/pep-0249/#exceptions
    assert helper().sql_exceptions == expected


@pytest.mark.parametrize('helper, expected', [
    (OracleDbHelper, (oracledb.DatabaseError, oracledb.InterfaceError)),
    (MSSQLDbHelper, (pyodbc.DatabaseError, pyodbc.InterfaceError)),
    (PostgresDbHelper, (psycopg2.DatabaseError, psycopg2.InterfaceError)),
    (SQLiteDbHelper, (sqlite3.DatabaseError, sqlite3.InterfaceError))
])
def test_connect_exceptions(helper, expected):
    # DBAPI 2.0 specifies that all exceptions must inherit from DatabaseError
    # or InterfaceError. See https://peps.python.org/pep-0249/#exceptions
    assert helper().connect_exceptions == expected


@pytest.mark.parametrize('helper, expected', [
    (OracleDbHelper, 'named'),
    (MSSQLDbHelper, 'qmark'),
    (PostgresDbHelper, 'pyformat'),
    (SQLiteDbHelper, 'qmark')
])
def test_paramstyle(helper, expected):
    assert helper().paramstyle == expected


@pytest.mark.parametrize('db_params, driver, expected', [
    (ORACLEDB, oracledb, 'testuser/mypassword@server:1521/testdb'),
    (MSSQLDB, pyodbc,
     'DRIVER=test driver;SERVER=tcp:server;PORT=1521;DATABASE=testdb;UID=testuser;PWD=mypassword'), # NOQA
    (POSTGRESDB, psycopg2,
     'host=server port=1521 dbname=testdb user=testuser password=mypassword'),
    (SQLITEDB, sqlite3, '/myfile.db')
])
def test_connect(monkeypatch, db_params, driver, expected):
    # Arrange
    monkeypatch.setenv('DB_PASSWORD', 'mypassword')
    mock_connect = Mock()
    monkeypatch.setattr(driver, 'connect', mock_connect)
    DB_HELPER_FACTORY.from_dbtype.cache_clear()
    helper = DB_HELPER_FACTORY.from_db_params(db_params)

    # Act
    helper.connect(db_params, 'DB_PASSWORD')

    # Assert
    mock_connect.assert_called_with(expected)


@pytest.mark.parametrize('db_params, expected', [
    (ORACLEDB, 'oracle+oracledb://testuser:mypassword@server:1521/testdb'),
    (MSSQLDB,
     'mssql+pyodbc://testuser:mypassword@server:1521/testdb?driver=test+driver'),
    # NOQA
    (POSTGRESDB, 'postgresql://testuser:mypassword@server:1521/testdb'),
    (SQLITEDB, 'sqlite:////myfile.db')
])
def test_sqlalchemy_conn_string(monkeypatch, db_params, expected):

    monkeypatch.setenv('DB_PASSWORD', 'mypassword')
    helper = DB_HELPER_FACTORY.from_db_params(db_params)
    conn_str = helper.get_sqlalchemy_connection_string(db_params, 'DB_PASSWORD')

    assert conn_str == expected


@pytest.mark.parametrize('db_params, driver', [
    (ORACLEDB, 'oracledb'), (MSSQLDB, 'pyodbc'), (POSTGRESDB, 'psycopg2'),
    (SQLITEDB, 'sqlite3')
])
def test_connect_without_driver_raises_exception(db_params, driver, monkeypatch):
    """
    Test that exception and helpful message are raised when connecting
    without driver installed
    """
    # Arrange - simulate missing driver module.
    # See https://stackoverflow.com/a/2481588/3508733
    real_import = builtins.__import__

    def raise_import_error(name, globals, locals, fromlist, level):
        if name == driver:
            raise ImportError()
        return real_import(name, globals, locals, fromlist, level)

    monkeypatch.setattr(builtins, '__import__', raise_import_error)
    monkeypatch.setenv("PASSWORD_VARIABLE", "blahblahblah")
    DB_HELPER_FACTORY.from_dbtype.cache_clear()

    # Act and assert
    with pytest.warns(UserWarning), pytest.raises(ETLHelperConnectionError) as excinfo:
        db_params.connect('PASSWORD_VARIABLE')

    # Confirm error message includes driver details
    error_message = excinfo.value.args[0]
    assert "Could not import" in error_message
    assert driver in error_message


@pytest.mark.parametrize('db_params', [ORACLEDB, MSSQLDB, POSTGRESDB])
def test_connect_without_password_variable_raises_exception(db_params):
    with pytest.raises(ETLHelperConnectionError,
                       match=r"Name of password environment variable .* is required"):
        db_params.connect(None)


@pytest.mark.parametrize('db_params', [ORACLEDB, MSSQLDB, POSTGRESDB])
def test_connect_with_unset_password_variable_raises_exception(db_params):
    with pytest.raises(ETLHelperConnectionError,
                       match=r"Password environment variable .* is not set"):
        db_params.connect("This environment variable is not set")



"""ETL Helper script to demonstrate autogenerated field values being returned."""
import etlhelper as etl
from my_databases import POSTGRESDB

insert_sql = "INSERT INTO my_table (message) VALUES ('hello') RETURNING id"

with POSTGRESDB.connect("PGPASSWORD") as conn:
    result = etl.fetchone(insert_sql, conn)

print(result["id"])



"""ETL Helper script to demonstrate copy_table_rows."""
import etlhelper as etl
from my_databases import POSTGRESDB, ORACLEDB

with ORACLEDB.connect("ORA_PASSWORD") as src_conn:
    with POSTGRESDB.connect("PG_PASSWORD") as dest_conn:
        etl.copy_table_rows("my_table", src_conn, dest_conn)



"""ETL Helper script to demonstrate oracle handling Large Objects (LOBs)."""
import etlhelper as etl
import oracledb
from my_databases import ORACLEDB

select_sql = "SELECT my_clob, my_blob FROM my_table"

with ORACLEDB.connect("ORA_PASSWORD") as conn:
    # By default, ETL Helper returns native types
    result_as_native = etl.fetchall(select_sql, conn)

    # Update oracledb settings to return LOBs
    oracledb.defaults.fetch_lobs = True
    result_as_lobs = etl.fetchall(select_sql, conn)



"""Test for the helper factory"""
from unittest.mock import (
    MagicMock,
    Mock,
)

import oracledb
import psycopg2
import pyodbc
import pytest

from etlhelper import DbParams
from etlhelper.db_helper_factory import DB_HELPER_FACTORY
from etlhelper.exceptions import ETLHelperHelperError
from etlhelper.db_helpers import OracleDbHelper, PostgresDbHelper, MSSQLDbHelper


@pytest.mark.parametrize("dbtype_keyword, expected_helper",
                         [('ORACLE', OracleDbHelper),
                          ('PG', PostgresDbHelper),
                          ('MSSQL', MSSQLDbHelper)])
def test_from_dbparams(dbtype_keyword, expected_helper):
    """
    Tests correct helper produced given a db params object
    """
    db_params = MagicMock(DbParams)
    db_params.dbtype = dbtype_keyword
    helper = DB_HELPER_FACTORY.from_db_params(db_params)
    assert isinstance(helper, expected_helper)


@pytest.mark.parametrize("expected_helper, db_class",
                         [(OracleDbHelper, oracledb.Connection),
                          (PostgresDbHelper, psycopg2.extensions.connection),
                          (MSSQLDbHelper, pyodbc.Connection)])
def test_from_conn(expected_helper, db_class):
    """
    Tests correct helper produced given a conn object
    """
    conn = Mock()
    # conn.__class__ = oracledb.Connection
    conn.__class__ = db_class
    helper = DB_HELPER_FACTORY.from_conn(conn)
    assert isinstance(helper, expected_helper)


def test_from_conn_not_registered():
    """
    Tests helpful error message on attempt to choose unregistered conn type.
    """
    conn = Mock()
    conn.__class__ = "Not a real class"

    with pytest.raises(ETLHelperHelperError,
                       match=r'Unsupported connection type.*'):
        DB_HELPER_FACTORY.from_conn(conn)


def test_from_db_params_not_registered():
    """
    Tests helpful error message on attempt to choose unregistered db_params
    type.
    """
    db_params = MagicMock(DbParams)
    db_params.dbtype = 'Not a real type'

    with pytest.raises(ETLHelperHelperError,
                       match=r'Unsupported DbParams.dbtype.*'):
        DB_HELPER_FACTORY.from_db_params(db_params)


def test_from_conn_bad_type():
    with pytest.raises(ETLHelperHelperError,
                       match=r'Expected connection-like object.*'):
        DB_HELPER_FACTORY.from_conn('some string')


def test_from_db_params_bad_type():
    with pytest.raises(ETLHelperHelperError,
                       match=r'Expected DbParams-like object.*'):
        DB_HELPER_FACTORY.from_db_params('some string')



"""Integration tests for MS SQL Server database.
These currently run against internal BGS instance.
"""
# pylint: disable=unused-argument, missing-docstring
import os
from collections import namedtuple
from textwrap import dedent
from unittest.mock import Mock

import pyodbc
import pytest

from etlhelper import (
    DbParams,
    connect,
    copy_rows,
    copy_table_rows,
    execute,
    executemany,
    fetchall,
    generate_insert_sql,
    load,
)
from etlhelper.utils import (
    table_info,
    Column,
)
from etlhelper.db_helper_factory import DB_HELPER_FACTORY
from etlhelper.exceptions import (
    ETLHelperConnectionError,
    ETLHelperInsertError,
    ETLHelperQueryError
)
from etlhelper.row_factories import namedtuple_row_factory

# Skip these tests if database is unreachable
if not os.getenv('TEST_MSSQL_HOST'):
    pytest.skip('MSSQL test database is not defined', allow_module_level=True)

MSSQLDB = DbParams.from_environment(prefix='TEST_MSSQL_')
if not MSSQLDB.is_reachable():
    pytest.skip('MSSQL test database is unreachable', allow_module_level=True)


# -- Tests here --

def test_connect():
    conn = connect(MSSQLDB, 'TEST_MSSQL_PASSWORD',
                   trust_server_certificate=True)
    assert isinstance(conn, pyodbc.Connection)


@pytest.mark.parametrize('trust_server_certificate', [
    True,
    False
])
def test_connect_trust_server_certificate(monkeypatch, trust_server_certificate):
    # Arrange
    helper = DB_HELPER_FACTORY.from_db_params(MSSQLDB)
    mock_connect = Mock()
    monkeypatch.setattr(helper, '_connect_func', mock_connect)

    # Act
    helper.connect(MSSQLDB, 'TEST_MSSQL_PASSWORD',
                   trust_server_certificate=trust_server_certificate)

    # Assert
    conn_str = mock_connect.call_args.args[0]
    trusted = ';TrustServerCertificate=yes' in conn_str
    assert trust_server_certificate is trusted


def test_connect_wrong_password(monkeypatch):
    monkeypatch.setitem(os.environ, 'TEST_MSSQL_PASSWORD', 'bad_password')
    with pytest.raises(ETLHelperConnectionError):
        connect(MSSQLDB, 'TEST_MSSQL_PASSWORD')


def test_bad_select(testdb_conn):
    select_sql = "SELECT * FROM bad_table"
    with pytest.raises(ETLHelperQueryError):
        execute(select_sql, testdb_conn)


def test_bad_insert(testdb_conn):
    insert_sql = "INSERT INTO bad_table (id) VALUES (1)"
    with pytest.raises(ETLHelperQueryError):
        execute(insert_sql, testdb_conn)


def test_bad_constraint(test_tables, testdb_conn):
    # src already has a row with id=1
    insert_sql = "INSERT INTO src (id) VALUES (1)"
    with pytest.raises(ETLHelperQueryError):
        execute(insert_sql, testdb_conn)


def test_copy_rows_happy_path_fast_true(
        test_tables, testdb_conn, testdb_conn2, test_table_data_dict):
    # Note: ODBC driver requires separate connections for source and destination,
    # even if they are the same database.
    # Arrange and act
    select_sql = "SELECT * FROM src"
    insert_sql = INSERT_SQL.format(tablename='dest')
    copy_rows(select_sql, testdb_conn, insert_sql, testdb_conn2,
              row_factory=namedtuple_row_factory)

    # Assert
    sql = "SELECT * FROM dest"
    result = fetchall(sql, testdb_conn)
    assert result == test_table_data_dict


@pytest.mark.skip(reason="Since MSODBC18 this test is unreliable")
def test_copy_rows_happy_path_deprecated_tables_fast_true(
        test_deprecated_tables, testdb_conn, testdb_conn2, test_table_data_dict):
    # Note: ODBC driver requires separate connections for source and destination,
    # even if they are the same database.
    # Arrange and act
    select_sql = "SELECT * FROM src"
    insert_sql = INSERT_SQL.format(tablename='dest')
    with pytest.warns(UserWarning) as record:
        copy_rows(select_sql, testdb_conn, insert_sql, testdb_conn2,
                  row_factory=namedtuple_row_factory)

    # Assert
    assert len(record) == 1
    assert str(record[0].message).startswith(
        "fast_executemany execution failed")

    sql = "SELECT * FROM dest"
    result = fetchall(sql, testdb_conn)
    assert result == test_table_data_dict


def test_copy_rows_happy_path_fast_false(
        test_tables, testdb_fast_false_conn, testdb_fast_false_conn2, test_table_data_dict):
    # Note: ODBC driver requires separate connections for source and destination,
    # even if they are the same database.
    # Arrange and act
    select_sql = "SELECT * FROM src"
    insert_sql = INSERT_SQL.format(tablename='dest')
    copy_rows(select_sql, testdb_fast_false_conn, insert_sql, testdb_fast_false_conn2,
              row_factory=namedtuple_row_factory)

    # Assert
    sql = "SELECT * FROM dest"
    result = fetchall(sql, testdb_fast_false_conn)
    assert result == test_table_data_dict


def test_copy_rows_happy_path_deprecated_tables_fast_false(
        test_deprecated_tables, testdb_fast_false_conn,
        testdb_fast_false_conn2, test_table_data_dict):
    # Note: ODBC driver requires separate connections for source and destination,
    # even if they are the same database.
    # Arrange and act
    select_sql = "SELECT * FROM src"
    insert_sql = INSERT_SQL.format(tablename='dest')
    copy_rows(select_sql, testdb_fast_false_conn, insert_sql, testdb_fast_false_conn2,
              row_factory=namedtuple_row_factory)

    # Assert
    sql = "SELECT * FROM dest"
    result = fetchall(sql, testdb_fast_false_conn)
    assert result == test_table_data_dict


def test_copy_table_rows_happy_path_fast_true(
        test_tables, testdb_conn, testdb_conn2, test_table_data_dict):
    # Note: ODBC driver requires separate connections for source and destination,
    # even if they are the same database.
    # Arrange and act
    copy_table_rows('src', testdb_conn, testdb_conn2, target='dest',
                    row_factory=namedtuple_row_factory)

    # Assert
    sql = "SELECT * FROM dest"
    result = fetchall(sql, testdb_conn)
    assert result == test_table_data_dict


def test_copy_table_rows_on_error(test_tables, testdb_conn, test_table_data_dict):
    # Arrange
    duplicate_id_row_sql = """
       INSERT INTO dest (id, value)
       VALUES (
         1,
         1.234
         )""".strip()
    execute(duplicate_id_row_sql, testdb_conn)

    # Act
    errors = []
    copy_table_rows('src', testdb_conn, testdb_conn, target='dest',
                    on_error=errors.extend, row_factory=namedtuple_row_factory)

    # Assert
    sql = "SELECT * FROM dest"
    result = fetchall(sql, testdb_conn)

    # Check that first row was caught as error
    row, exception = errors[0]
    assert row.id == 1
    assert "unique" in str(exception).lower()

    # Check that other rows were inserted correctly
    assert result[1:] == test_table_data_dict[1:]


def test_fetchall_with_parameters(test_tables, testdb_conn):
    # parameters=None is tested by default in other tests

    # Bind by index
    sql = "SELECT * FROM src where ID = ?"
    result = fetchall(sql, testdb_conn, parameters=(1,))
    assert len(result) == 1
    assert result[0]["id"] == 1


def test_copy_rows_bad_param_style(test_tables, testdb_conn):
    # Arrange and act
    select_sql = "SELECT * FROM src"
    insert_sql = BAD_PARAM_STYLE_SQL.format(tablename='dest')
    with pytest.raises(ETLHelperInsertError):
        copy_rows(select_sql, testdb_conn, insert_sql, testdb_conn,
                  row_factory=namedtuple_row_factory)


def test_executemany_dicts_raises_error(testdb_conn, test_tables, test_table_data_dict):
    # Arrange
    # Placeholder doesn't really matter as pydodbc doesn't support
    # named placeholders
    insert_sql = dedent("""
        INSERT INTO dest (id, value, simple_text, utf8_text, day, date_time)
        VALUES (:id, :value, :simple_text, :utf8_text, :day, :date_time)
        ;""").strip()
    expected_message = ("pyodbc driver for MS SQL only supports positional placeholders.  "
                        "Use namedtuple, tuple or list (via row_factory setting for copy_rows).")

    # Act and assert
    # pyodbc doesn't support named parameters.
    with pytest.raises(ETLHelperInsertError) as exc_info:
        executemany(insert_sql, testdb_conn, test_table_data_dict)

    assert str(exc_info.value) == expected_message


def test_load_namedtuples(testdb_conn, test_tables, test_table_data_namedtuple):
    # Act
    load('dest', testdb_conn, test_table_data_namedtuple)

    # Assert
    sql = "SELECT * FROM dest"
    result = fetchall(sql, testdb_conn, row_factory=namedtuple_row_factory)
    assert result == test_table_data_namedtuple


def test_load_dicts(testdb_conn, test_tables, test_table_data_dict):
    # Arrange
    expected_message = ("Database connection (<class 'pyodbc.Connection'>) doesn't support named parameters.  "
                        "Pass data as namedtuples instead.")

    # Act and assert
    # pyodbc doesn't support named parameters.
    with pytest.raises(ETLHelperInsertError) as exc_info:
        load('dest', testdb_conn, test_table_data_dict)

    assert str(exc_info.value) == expected_message


def test_generate_insert_sql_tuple(testdb_conn):
    # Act
    data = (1, 'one')
    with pytest.raises(ETLHelperInsertError,
                       match="Row is not a dictionary or namedtuple"):
        generate_insert_sql('my_table', data, testdb_conn)


def test_generate_insert_sql_namedtuple(testdb_conn):
    # Arrange
    TwoColumnRow = namedtuple('TwoColumnRow', ('id', 'data'))
    data = TwoColumnRow(1, 'one')
    expected = 'INSERT INTO my_table (id, data) VALUES (?, ?)'

    # Act
    sql = generate_insert_sql('my_table', data, testdb_conn)

    # Assert
    assert sql == expected


def test_generate_insert_sql_dictionary(testdb_conn):
    # Act
    data = {'id': 1, 'data': 'one'}
    with pytest.raises(ETLHelperInsertError,
                       match="doesn't support named parameters"):
        generate_insert_sql('my_table', data, testdb_conn)


def test_table_info_no_schema_no_duplicates(testdb_conn, test_tables):
    # Arrange
    expected = [
        Column(name='id', type='int', not_null=0, has_default=0),
        Column(name='value', type='float', not_null=1, has_default=0),
        Column(name='simple_text', type='nvarchar', not_null=0, has_default=1),
        Column(name='utf8_text', type='nvarchar', not_null=0, has_default=0),
        Column(name='day', type='date', not_null=0, has_default=0),
        Column(name='date_time', type='datetime2', not_null=0, has_default=0)
    ]

    # Act
    columns = table_info('src', testdb_conn)

    # Assert
    assert columns == expected


def test_table_info_with_schema_no_duplicates(testdb_conn, test_tables):
    # Arrange
    expected = [
        Column(name='id', type='int', not_null=0, has_default=0),
        Column(name='value', type='float', not_null=1, has_default=0),
        Column(name='simple_text', type='nvarchar', not_null=0, has_default=1),
        Column(name='utf8_text', type='nvarchar', not_null=0, has_default=0),
        Column(name='day', type='date', not_null=0, has_default=0),
        Column(name='date_time', type='datetime2', not_null=0, has_default=0)
    ]

    # Act
    columns = table_info('src', testdb_conn, schema='etlhelper')

    # Assert
    assert columns == expected


def test_table_info_bad_table_name_no_schema(testdb_conn, test_tables):
    # Arrange, act and assert
    with pytest.raises(ETLHelperQueryError, match=r"Table name 'bad_table' not found."):
        table_info('bad_table', testdb_conn)


def test_table_info_bad_table_name_with_schema(testdb_conn, test_tables):
    # Arrange, act and assert
    with pytest.raises(ETLHelperQueryError, match=r"Table name 'etlhelper.bad_table' not found."):
        table_info('bad_table', testdb_conn, schema='etlhelper')


# -- Fixtures here --

INSERT_SQL = dedent("""
    INSERT INTO {tablename} (id, value, simple_text, utf8_text,
      day, date_time)
    VALUES
      (?, ?, ?, ?, ?, ?)
      ;""").strip()

BAD_PARAM_STYLE_SQL = dedent("""
    INSERT INTO {tablename} (id, value, simple_text, utf8_text,
      day, date_time)
    VALUES
      (%s, %s, %s, %s, %s, %s)
      """).strip()


@pytest.fixture(scope='function')
def testdb_conn():
    """Get connection to test MS SQL database."""
    with connect(MSSQLDB, 'TEST_MSSQL_PASSWORD',
                 trust_server_certificate=True) as conn:
        return conn


@pytest.fixture(scope='function')
def testdb_conn2():
    """Get connection to test MS SQL database."""
    with connect(MSSQLDB, 'TEST_MSSQL_PASSWORD',
                 trust_server_certificate=True) as conn:
        return conn


@pytest.fixture(scope='function')
def testdb_fast_false_conn():
    """Get connection to test MS SQL database."""
    with connect(MSSQLDB, 'TEST_MSSQL_PASSWORD', fast_executemany=False,
                 trust_server_certificate=True) as conn:
        return conn


@pytest.fixture(scope='function')
def testdb_fast_false_conn2():
    """Get connection to test MS SQL database."""
    with connect(MSSQLDB, 'TEST_MSSQL_PASSWORD', fast_executemany=False,
                 trust_server_certificate=True) as conn:
        return conn


@pytest.fixture(scope='function')
def test_tables(test_table_data_namedtuple, testdb_conn):
    """
    Create a table and fill with test data.  Teardown after the yield drops it
    again.
    """
    # Define SQL queries
    drop_src_sql = "DROP TABLE src"
    create_src_sql = dedent("""
        CREATE TABLE src
          (
            id integer unique,
            value double precision not null,
            simple_text nvarchar(max) default 'default',
            utf8_text nvarchar(max),
            day date,
            date_time datetime2(6)
          )
          ;""").strip()
    drop_dest_sql = drop_src_sql.replace('src', 'dest')
    create_dest_sql = create_src_sql.replace('src', 'dest')

    # Create table and populate with test data
    with testdb_conn.cursor() as cursor:
        # src table
        try:
            cursor.execute(drop_src_sql)
        except pyodbc.DatabaseError:
            pass
        cursor.execute(create_src_sql)
        cursor.executemany(INSERT_SQL.format(tablename='src'),
                           test_table_data_namedtuple)
        # dest table
        try:
            cursor.execute(drop_dest_sql)
        except pyodbc.DatabaseError:
            # Error if table doesn't exist
            pass
        cursor.execute(create_dest_sql)
    testdb_conn.commit()

    # Return control to calling function until end of test
    yield

    # Tear down the table after test completes
    with testdb_conn.cursor() as cursor:
        cursor.execute(drop_src_sql)
        cursor.execute(drop_dest_sql)
    testdb_conn.commit()


@pytest.fixture(scope='function')
def test_deprecated_tables(test_table_data_namedtuple, testdb_conn):
    """
    Create a table and fill with test data.  Teardown after the yield drops it
    again.
    """
    # Define SQL queries
    drop_src_sql = "DROP TABLE src"
    create_src_sql = dedent("""
        CREATE TABLE src
          (
            id integer unique,
            value double precision,
            simple_text text,
            utf8_text text,
            day date,
            date_time datetime2(6)
          )
          ;""").strip()
    drop_dest_sql = drop_src_sql.replace('src', 'dest')
    create_dest_sql = create_src_sql.replace('src', 'dest')

    # Create table and populate with test data
    with testdb_conn.cursor() as cursor:
        # src table
        try:
            cursor.execute(drop_src_sql)
        except pyodbc.DatabaseError:
            pass
        cursor.execute(create_src_sql)
        cursor.executemany(INSERT_SQL.format(tablename='src'),
                           test_table_data_namedtuple)
        # dest table
        try:
            cursor.execute(drop_dest_sql)
        except pyodbc.DatabaseError:
            # Error if table doesn't exist
            pass
        cursor.execute(create_dest_sql)
    testdb_conn.commit()

    # Return control to calling function until end of test
    yield

    # Tear down the table after test completes
    with testdb_conn.cursor() as cursor:
        cursor.execute(drop_src_sql)
        cursor.execute(drop_dest_sql)
    testdb_conn.commit()



from typing import (
    Any,
    Collection,
    Protocol,
)


class Connection(Protocol):
    def close(self) -> None:
        ...

    def commit(self) -> None:
        ...

    def rollback(self) -> None:
        ...

    def cursor(self):  # noqa Cursor Protocol not defined
        ...


Row = Collection[Any]
Chunk = list[Row]



"""
ETLHelper exceptions are derived from the ETLHelperError base class.

"""


class ETLHelperError(Exception):
    """Base class for exceptions in this module"""


class ETLHelperConnectionError(ETLHelperError):
    """Exception raised for bad database connections"""


class ETLHelperQueryError(ETLHelperError):
    """Exception raised for SQL query errors and similar"""


class ETLHelperDbParamsError(ETLHelperError):
    """Exception raised for bad database parameters"""


class ETLHelperExtractError(ETLHelperError):
    """Exception raised when extracting data."""


class ETLHelperInsertError(ETLHelperError):
    """Exception raised when inserting data."""


class ETLHelperAbort(ETLHelperError):
    """Exception raised when `abort` is called."""


class ETLHelperHelperError(ETLHelperError):
    """Exception raised when helper selection fails."""


class ETLHelperBadIdentifierError(ETLHelperError):
    """
    Exception raised when identifier contains invalid characters.  This may
    indicate an attempt at SQL injection.
    """



"""
Connect to database
"""
from etlhelper.db_helper_factory import DB_HELPER_FACTORY


def connect(db_params, password_variable=None, **kwargs):
    """
    Return database connection.

    :param db_params: DbParams object or similar with appropriate attributes
    :param password_variable: str, name of environment variable with password
    :param kwargs: connection specific keyword arguments e.g. row_factory
    :return: Connection object
    """
    helper = DB_HELPER_FACTORY.from_db_params(db_params)
    # Helpers will raise ETL HelperConnectionError if connection fails
    conn = helper.connect(db_params, password_variable, **kwargs)
    return conn


def get_connection_string(db_params, password_variable):
    """
    Get a connection string

    :param db_params: DbParams object or similar with appropriate attributes
    :param password_variable: str, name of environment variable with password
    :return: str, Connection string
    """
    helper = DB_HELPER_FACTORY.from_db_params(db_params)
    return helper.get_connection_string(db_params, password_variable)


def get_sqlalchemy_connection_string(db_params, password_variable):
    """
    Get a SQLAlchemy connection string.

    :param db_params: DbParams object or similar with appropriate attributes
    :param password_variable: str, name of environment variable with password
    :return: str, Connection string
    """
    helper = DB_HELPER_FACTORY.from_db_params(db_params)
    return helper.get_sqlalchemy_connection_string(db_params, password_variable)



"""
Test db params
"""
from unittest.mock import Mock

import pytest

from etlhelper.db_params import DbParams
from etlhelper.exceptions import ETLHelperDbParamsError
import etlhelper.db_params as etl_db_params


@pytest.fixture(scope='function')
def test_params():
    """Return an example DbParams class."""
    test_params_example = DbParams(
        dbtype='PG',
        host='localhost',
        port=5432,
        dbname='etlhelper',
        user='etlhelper_user')
    return test_params_example


def test_db_params_validate_params_invalid_dbtype():
    with pytest.raises(ETLHelperDbParamsError, match=r'.* not in valid types .*') as excinfo:
        DbParams(dbtype='not valid')

    # Check that exception doesn't include lower exceptions from stack
    exception_chain = excinfo.getrepr().chain
    assert len(exception_chain) == 1


def test_db_params_validate_params_missing_params():
    with pytest.raises(ETLHelperDbParamsError,
                       match=r"{'user'} not set. Required parameters are .*"):
        # missing user
        DbParams(dbtype='ORACLE', host='localhost', port=1, dbname='test')


def test_db_params_validate_params_extra_param():
    with pytest.raises(ETLHelperDbParamsError,
                       match=r"Invalid parameter\(s\): {'extra_param'}"):
        DbParams(dbtype='ORACLE', host='localhost', port=1, dbname='test',
                 user='user', extra_param=1)


def test_db_params_repr(test_params):
    """Test DbParams string representation"""
    result = str(test_params)
    expected = ("DbParams(host='localhost', "
                "port='5432', dbname='etlhelper', "
                "user='etlhelper_user', dbtype='PG')")
    assert result == expected


def test_db_params_setattr(test_params):
    """Test that we can set a DbParams attribute using a __setattr__ approach"""
    # Set a param using dot notation
    test_params.user = "Data McDatabase"

    assert test_params.user == "Data McDatabase"


def test_db_params_setattr_bad_param(test_params):
    """Test that __setattr__ approach fails for bad parameter"""
    with pytest.raises(ETLHelperDbParamsError, match=r"Invalid parameter\(s\): {'some_bad_param'}"):
        # Set a param using dot notation and confirm error
        test_params.some_bad_param = "Data McDatabase"
    # Confirm param is not set
    assert test_params.get('some_bad_param', 'not_found') == 'not_found'


def test_db_params_from_environment(monkeypatch):
    """
    Test capturing db params from environment settings.
    """
    # Arrange
    monkeypatch.setenv('TEST_DB_PARAMS_ENV_DBTYPE', 'ORACLE')
    monkeypatch.setenv('TEST_DB_PARAMS_ENV_HOST', 'test.host')
    monkeypatch.setenv('TEST_DB_PARAMS_ENV_PORT', '1234')
    monkeypatch.setenv('TEST_DB_PARAMS_ENV_DBNAME', 'testdb')
    monkeypatch.setenv('TEST_DB_PARAMS_ENV_USER', 'testuser')
    monkeypatch.setenv('TEST_DB_PARAMS_ENV_PASSWORD', 'dont want this')

    # Act
    db_params = DbParams.from_environment(prefix='TEST_DB_PARAMS_ENV_')

    # Assert
    assert db_params.dbtype == 'ORACLE'
    assert db_params.host == 'test.host'
    assert db_params.port == '1234'
    assert db_params.dbname == 'testdb'
    assert db_params.user == 'testuser'

    # Confirm that password variable wasn't included
    assert 'PASSWORD' not in db_params
    assert 'password' not in db_params


def test_db_params_from_environment_not_set(monkeypatch):
    """
    Test missing db params from environment settings.
    """
    # Arrange
    monkeypatch.delenv('TEST_DBTYPE', raising=False)

    # Act
    with pytest.raises(ETLHelperDbParamsError,
                       match=r".*environment variable is not set.*"):
        DbParams.from_environment(prefix='TEST_')


def test_db_params_copy(test_params):
    """
    Test db_params can copy themselves.
    """
    # Act
    test_params2 = test_params.copy()

    # Assert
    assert test_params2 == test_params
    assert test_params2 is not test_params
    assert isinstance(test_params2, DbParams)


def test_db_params_connect(test_params, monkeypatch):
    """
    Assert that the connect function is called with the params object
    and the password variable.
    """
    # Arrange
    mock_connect = Mock()
    monkeypatch.setattr(etl_db_params, "connect", mock_connect)

    # Act
    test_params.connect("PASSWORD_VARIABLE", foobarkey="blah")

    # Assert
    mock_connect.assert_called_once_with(test_params, "PASSWORD_VARIABLE", foobarkey="blah")


def test_db_params_get_connection_string(test_params, monkeypatch):
    """
    Test that the correct connection string is given when using the
    DbParams method for connect.
    """
    # Arrange
    monkeypatch.setenv("PASSWORD_VARIABLE", "blahblahblah")

    # Act
    conn_str = test_params.get_connection_string("PASSWORD_VARIABLE")

    # Assert
    assert conn_str == 'host=localhost port=5432 dbname=etlhelper user=etlhelper_user password=blahblahblah'  # noqa


def test_db_params_get_sqlalchemy_connection_string(test_params, monkeypatch):
    """
    Test that the correct connection string is given when using the
    DbParams method for connect, for a SQL Alchemy.
    """
    # Arrange
    monkeypatch.setenv("PASSWORD_VARIABLE", "blahblahblah")

    # Act
    conn_str = test_params.get_sqlalchemy_connection_string("PASSWORD_VARIABLE")

    # Assert
    assert conn_str == 'postgresql://etlhelper_user:blahblahblah@localhost:5432/etlhelper'


def test_db_params_paramstyle(test_params):
    """Test that the correct paramstyle string is returned."""
    assert test_params.paramstyle == 'pyformat'



"""
Script to create database and load observations data from csv file. It also
demonstrates how an `on_error` function can handle failed rows.

Generate observations.csv with:
curl 'https://sensors.bgs.ac.uk/FROST-Server/v1.1/Observations?$select=@iot.id,result,phenomenonTime&$top=20000&$resultFormat=csv' -o observations.csv  # noqa
"""
import csv
import sqlite3
import datetime as dt
from typing import Iterable

import etlhelper as etl


def load_observations(csv_file: str, conn: sqlite3.Connection) -> None:
    """Load observations from csv_file to db_file."""
    # Drop table (helps with repeated test runs!)
    drop_table_sql = """
        DROP TABLE IF EXISTS observations
        """
    etl.execute(drop_table_sql, conn)

    # Create table (reject ids with no remainder when divided by 1000)
    create_table_sql = """
        CREATE TABLE IF NOT EXISTS observations (
            id INTEGER PRIMARY KEY CHECK (id % 1000),
            time TIMESTAMP,
            result FLOAT
            )"""
    etl.execute(create_table_sql, conn)

    # Load data
    with open(csv_file, "rt") as f:
        reader = csv.DictReader(f)
        etl.load("observations", conn, reader, transform=transform, on_error=on_error)


# A transform function that takes an iterable of rows and returns an iterable
# of rows.
def transform(rows: Iterable[dict]) -> Iterable[dict]:
    """Rename time column and convert to Python datetime."""
    for row in rows:
        # Dictionaries are mutable, so rows can be modified in place.
        time_value = row.pop("phenomenonTime")
        row["time"] = dt.datetime.strptime(time_value, "%Y-%m-%dT%H:%M:%S.%fZ")

    return rows


# The on_error function is called after each chunk with all the failed rows
def on_error(failed_rows: list[tuple[dict, Exception]]) -> None:
    """Print the IDs of failed rows"""
    rows, exceptions = zip(*failed_rows)
    failed_ids = [row["id"] for row in rows]
    print(f"Failed IDs: {failed_ids}")


if __name__ == "__main__":
    etl.log_to_console()

    db = etl.DbParams(dbtype="SQLITE", filename="observations.sqlite")
    with db.connect() as conn:
        load_observations("observations.csv", conn)



"""
Utility functions to help with tasks such as programatically generating SQL queries.
"""
from collections import namedtuple

from etlhelper import fetchall
from etlhelper.exceptions import ETLHelperQueryError
from etlhelper.db_helper_factory import DB_HELPER_FACTORY
from etlhelper.row_factories import namedtuple_row_factory

Column = namedtuple('Column', ['name', 'type', 'not_null', 'has_default'])


def table_info(table, conn, schema=None):
    """
    Return basic metadata for each of the columns of 'table' on 'conn'.

    :param table: str, the table to describe
    :param conn: dbapi connection
    :param schema: str, optional name of schema for table
    :returns columns: list, tuples of (name, type, not_null, has_default)
    """
    helper = DB_HELPER_FACTORY.from_conn(conn)

    params = (table, schema)
    result = fetchall(helper.table_info_query, conn, parameters=params, row_factory=namedtuple_row_factory)
    columns = [Column(*row) for row in result]

    if not columns:
        schema_table = f"{schema}.{table}" if schema else table
        msg = f"Table name '{schema_table}' not found."
        raise ETLHelperQueryError(msg)

    # If same table exists in another schema, duplicate columns may be returned
    if len(columns) > len(set(col.name for col in columns)):
        msg = (f"Table name {table} is not unique in database.  "
               "Please specify the schema.")
        raise ETLHelperQueryError(msg)

    return columns



"""
Library to simplify data transfer between databases
"""
import logging
import sys
from importlib.metadata import (
    PackageNotFoundError,
    version,
)
from typing import TextIO

# Import helper functions here for more convenient access
from etlhelper.abort import abort_etlhelper_threads
from etlhelper.db_params import DbParams
from etlhelper.etl import (
    copy_rows,
    execute,
    executemany,
    fetchone,
    fetchall,
    generate_insert_sql,
    iter_chunks,
    iter_rows,
    load,
    copy_table_rows,
)
from etlhelper.connect import (
    connect,
    get_connection_string,
    get_sqlalchemy_connection_string,
)
from etlhelper.utils import (
    table_info,
)
from etlhelper import (
    row_factories,
)

try:
    __version__ = version(__name__)
except PackageNotFoundError:
    __version__ = "0.0.0"

# Create etlhelper logger and clear the logger handlers
# This prevents a new logger from being created when running 'logging.getLogger("etlhelper")'
# with default handlers
logging.getLogger("etlhelper").handlers.clear()


def log_to_console(
    level: int = logging.INFO,
    output: TextIO = sys.stderr,
) -> None:
    """
    Log ETL Helper messages to the given output.

    :param level: logger level
    :param output: the output location of the logger messages
    """
    logger = logging.getLogger('etlhelper')
    # Clear all existing handlers to prevent duplicate output
    logger.handlers.clear()

    # Prepare log handler.  See this StackOverflow answer for details:
    # https://stackoverflow.com/a/27835318/3508733
    class CleanDebugMessageFormatter(logging.Formatter):
        default_fmt = logging.Formatter('%(asctime)s %(funcName)s: %(message)s')
        debug_fmt = logging.Formatter('%(message)s')

        def format(self, record: logging.LogRecord) -> str:
            if record.levelno < logging.INFO:
                return self.debug_fmt.format(record)
            else:
                return self.default_fmt.format(record)

    handler = logging.StreamHandler(output)
    handler.setFormatter(CleanDebugMessageFormatter())

    # Configure logger
    logger.addHandler(handler)
    logger.setLevel(level=level)


__all__ = [
    "DbParams",
    "abort_etlhelper_threads",
    "connect",
    "copy_rows",
    "copy_table_rows",
    "execute",
    "executemany",
    "fetchall",
    "fetchone",
    "generate_insert_sql",
    "get_connection_string",
    "get_sqlalchemy_connection_string",
    "iter_chunks",
    "iter_rows",
    "load",
    "log_to_console",
    "row_factories",
    "table_info",
]



"""
Database helper for PostgreSQL
"""
import warnings
from textwrap import dedent

from etlhelper.db_helpers.db_helper import DbHelper


class PostgresDbHelper(DbHelper):
    """
    Postgres db helper class
    """
    table_info_query = dedent("""
        SELECT
            pg_attribute.attname AS name,
            pg_catalog.format_type(pg_attribute.atttypid, pg_attribute.atttypmod) AS type,
            (case when pg_attribute.attnotnull then 1 else 0 end) as not_null,
            (case when pg_attribute.atthasdef then 1 else 0 end) as has_default
        FROM
            pg_catalog.pg_attribute
        INNER JOIN
            pg_catalog.pg_class ON pg_class.oid = pg_attribute.attrelid
        INNER JOIN
            pg_catalog.pg_namespace ON pg_namespace.oid = pg_class.relnamespace
        WHERE
            pg_attribute.attnum > 0
            AND NOT pg_attribute.attisdropped
            AND pg_class.relname = %s
            AND pg_namespace.nspname ~ COALESCE(%s, '.*')
        ORDER BY
            attnum ASC;
        """).strip()

    def __init__(self):
        super().__init__()
        self.required_params = {'host', 'port', 'dbname', 'user'}
        self.missing_driver_msg = (
            "Could not import psycopg2 module required for PostgreSQL connections.  "
            "See https://github.com/BritishGeologicalSurvey/etlhelper for installation instructions")
        self.named_paramstyle = 'pyformat'
        self.positional_paramstyle = 'format'

        try:
            import psycopg2
            self.sql_exceptions = (psycopg2.DatabaseError,
                                   psycopg2.InterfaceError)
            self.connect_exceptions = (psycopg2.DatabaseError,
                                       psycopg2.InterfaceError)
            self.paramstyle = psycopg2.paramstyle
            self._connect_func = psycopg2.connect
        except ImportError:
            warnings.warn(self.missing_driver_msg)

    def get_connection_string(self, db_params, password_variable):
        """
        Return a connection string
        :param db_params: DbParams
        :param password_variable: str, password
        :return: str
        """
        # Prepare connection string
        password = self.get_password(password_variable)
        return (f'host={db_params.host} port={db_params.port} '
                f'dbname={db_params.dbname} '
                f'user={db_params.user} password={password}')

    def get_sqlalchemy_connection_string(self, db_params, password_variable):
        """
        Returns connection string for SQLAlchemy engine.
        """
        password = self.get_password(password_variable)
        return (f'postgresql://{db_params.user}:{password}@'
                f'{db_params.host}:{db_params.port}/{db_params.dbname}')

    @staticmethod
    def executemany(cursor, query, chunk):
        """
        Call execute_batch method for PostGres.

        :param cursor: Open database cursor.
        :param query: str, SQL query
        :param chunk: list, Rows of parameters.
        """
        # Here we use execute_batch to send multiple inserts to db at once.
        # This is faster than executemany() because it results in fewer db
        # calls.  execute_values() or preparing single statement with
        # mogrify() were not used because resulting input statement is less
        # clear and selective formatting of inputs for spatial vs non-spatial
        # tables adds significant code complexity.
        # See following for background:
        # https://github.com/psycopg/psycopg2/issues/491#issuecomment-276551038
        # https://www.compose.com/articles/formatted-sql-in-python-with-psycopgs-mogrify/
        from psycopg2.extras import execute_batch

        execute_batch(cursor, query, chunk, page_size=len(chunk))



"""ETL Helper script to demonstrate load."""
import sqlite3
import etlhelper as etl

db_file = "igneous_rocks.db"

rows = [
    {"name": "basalt", "grain_size": "fine"},
    {"name": "granite", "grain_size": "coarse"}
]

with sqlite3.connect(db_file) as conn:
    # Note that table must already exist
    processed, failed = etl.load("igneous_rock", conn, rows)



"""ETL Helper script to demonstrate using executemany with a named placeholder query."""
import sqlite3
import etlhelper as etl

db_file = "igneous_rocks.db"

# Insert query changes case and adds update_at column
insert_sql = """
    INSERT INTO igneous_rocks (name, grain_size, updated_at)
    VALUES (:name, UPPER(:grainsize), datetime('now'))
"""

rows = [
    {"name": "basalt", "grain_size": "fine"},
    {"name": "granite", "grain_size": "coarse"}
]

with sqlite3.connect(db_file) as conn:
    # Note that table must already exist
    processed, failed = etl.executemany(insert_sql, conn, rows)



"""ETL Helper script to copy records between databases."""
import sqlite3
import datetime as dt
from typing import Generator
import etlhelper as etl

igneous_db_file = "igneous_rocks.db"
rock_db_file = "rocks.db"

create_sql = """
    CREATE TABLE IF NOT EXISTS rock (
        id INTEGER PRIMARY KEY,
        name TEXT UNIQUE,
        category TEXT,
        last_update DATETIME
    )
"""
select_sql = "SELECT name FROM igneous_rock"


def transform(chunk: list[dict]) -> Generator[dict, None, None]:
    for row in chunk:
        row["category"] = "igneous"
        row["last_update"] = dt.datetime.now()
        yield row


etl.log_to_console()

with sqlite3.connect(igneous_db_file) as src:
    with sqlite3.connect(rock_db_file) as dest:
        # Create target table
        etl.execute(create_sql, dest)

        # Copy data
        rows = etl.iter_rows(select_sql, src, transform=transform)
        etl.load("rock", dest, rows)

        # Confirm transfer
        for row in etl.fetchall("SELECT * FROM rock", dest):
            print(row)



"""ETL Helper script to demonstrate compatibility when creating an SQLAlchemy connection."""
import pandas as pd
from sqlalchemy import create_engine

from my_databases import ORACLEDB

engine = create_engine(ORACLEDB.get_sqlalchemy_connection_string("ORACLE_PASSWORD"))

sql = "SELECT * FROM my_table"
df = pd.read_sql(sql, engine)
df.to_csv("my_data.csv", header=True, index=False, float_format="%.3f")



"""ETL Helper script to demonstrate an extract error."""
import sqlite3
import etlhelper as etl

db_file = "igneous_rocks.db"
select_sql = "SELECT * FROM bad_table"

with sqlite3.connect(db_file) as conn:
    rows = etl.fetchall(select_sql, conn)
